#include <stdio.h>			/* fprintf, printf, ... */
#include <stdlib.h>			/* malloc, free, atoi, exit, ... */
#include <string.h>     		/* strcpy, ... */

/* -------------- ATTENTION:
	Ce prgm utilise les definitions de types 
	et de masques de selection a champs variables
	et ne peut donc etre utilise qu'avec une version 
	de T-Linda superieure a 3.5
*/
#include "linda3.h"			/* in, out, ... */

/********** definition des nouveaux types */

typedef struct				/* paquet de donnees */
{
  char d[8] ;				/* donnees de calcul */
  short c[1] ;				/* compteurs de passage */
} PAQUET ;

/********** prototypes locaux */
void parametres(int argc, char **argv, int *nbpaquets, int *nprocessus, int *nbcycles) ;
PAQUET *palloc(int nbprocessus) ;


/********** fonctions annexes */

void parametres(int argc, char **argv, int *nbpaquets, int *nprocessus, int *nbcycles)
{
  if (argc != 4)
    {
       fprintf(stderr, "\nusage: %s npaquets nprocessus ncycles\n", argv[0]);
       exit(-1);
    }
  *nbpaquets 	= atoi(argv[1]);
  *nprocessus 	= atoi(argv[2]);
  *nbcycles 	= atoi(argv[3]);       
}

PAQUET *palloc(int nbprocessus)
{
  PAQUET *paquet;

  paquet = (PAQUET *)malloc( sizeof(PAQUET) + (nbprocessus-1)*sizeof(short) );
  if (paquet == NULL)
    {
      fprintf(stderr, "pas assez de memoire pour creer un paquet\n") ;
      exit(-2);
    }
  return paquet;
}

/********** programme principal */

int main(int c, char *v[])
{
  int np;			/* nombre de paquets de donnees */
  int ne;			/* nombre de processus de travail */
  int nc;			/* nombre de cycles de calcul */
  PAQUET *p;			/* un paquet de donnees */
  TupleType tt;			/* type pour les typles "paquet" */
  TupleSelector ts;		/* selecteur pour les typles "paquet" */
  int i, k;			/* quelques variables de travail */
  int param[4] ;		/* table des parametres a passer a Eval() */

  /**** initialisations ****/

  parametres(c, v, &np, &ne, &nc);	/* interpreter la ligne de commande */

  tt = DefineTupleType("paquet", "{c[8] %ds}",ne ); /* definit le type "paquet" */
  /* descripteur de type de la forme {c[8] sssss} */
  p = palloc(ne);			/* alloue la memoire pour un paquet */

  /**** lancement des calculs ****/
  /* remplissage table des parametres */
  param[0] = np ;
  param[1] = ne ;
  param[2] = nc ;
  /* creation des processus de calcul */
  for( param[3] = 0 ; param[3] < ne ; ++param[3] )
    Eval("scycle", &tt, 1, param, sizeof(param)) ;
    
  for (i = 0; i < ne; ++i)		/* initialisation des compteurs de passage */
    p->c[i] = 0;

  for (i = 0; i < np; ++i)		/* nombre de paquets a creer */
    {
      sprintf(p->d, "D%03d", i);	/* lire un paquet du cycle courant */
      k = out(tt, p);			/* depose le paquet dans le T.S. */
      printf("status d'ecriture paquet %s : %d\n", p->d, k);
    }

  /**** synthese des resultats ****/

  ts = CreateTupleSelector(tt, "{? %d.}", ne );	/* cree un selecteur {? ....} */

  for (i = 0; i < ne; ++i)		/* parametres reels de selection */
    p->c[i] = nc;

  for (i = 0; i < np; ++i)		/* nombre de paquets a reprendre */
    {
      k = in(ts, p);			/* retrait d'un paquet du T.S. */
      printf("status de lecture du paquet %s : %d\n", p->d, k);
    }

  /**** terminaisons ****/

  free(p);			/* destruction du paquet */
  free(ts);			/* destruction du selecteur */
  UndefTupleType(tt);		/* destruction du type */
}
