
/************************************************************************/
/*									*/
/*	lnd.c : Serveur de M.A	Version 3.0				*/
/*									*/
/*	Copyright (C) 1991-94 LGIL					*/
/*									*/
/************************************************************************/

#include "linda3.h"

#include <helios.h>
#include <string.h>
#include <task.h>
#include <signal.h>
#include <syslib.h>

/* definitions pouvant servir au debug
***************************************/
#define miss(xxx)\
{ fprintf(stdout,"\n[%s]",xxx) ; fflush(stdout) ; }
#define mess(xxx)
void      dump_types(void);
void      print_types(TUPLE_TYPE * t);


/* Liste des fonctions "utilisateur" du serveur
***********************************************/

typedef int (*IntFnPtr) ();

VoidFnPtr LindaFunction[] =
{
	LindaPutWord,
	LindaPut,
	LindaGet,
	LindaRead,
	LindaGetFast,
	LindaReadFast,
	LindaGetU,
	LindaReadU,
	LindaGetFastU,
	LindaReadFastU,
	LindaKill,
	LindaChangeDbgLevel,
	LindaNop,
	LindaSynchronize,
	LindaPrint,
	LindaPutType,
	LindaGetTypeFast,
	LindaOut,
	LindaIn,
	LindaRd
};

#define MaxLindaFunction (sizeof(LindaFunction)/sizeof(*LindaFunction))

/* Donnees privees
******************/

static byte servname[16];	/* Nom du serveur	*/
static word servnum = -1;	/* numero(-1 si inconnu)*/
static TupleSpace main_tps;	/* Espace de stockage	*/
static Port port;		/* port principal de    */

/* reception des requetes */
static int servdbg = 0;		/* 1 si serveur de debug */
static int dbg_level = 0;	/* != 0 si sous debug */
static int fifo = 1;		/* 0 si l'ordre de lecture
				** des tuples n'EST PAS leur
				** ordre d'ecriture 
				*/

/* point d'entree dans la structure de directory du serveur:
   ********************************************************

   ceci n'est utile que si on utilise le dispatcher standard de helios
   (Dispatch)
   Cette structure est tres imcomplete: Juste un directory.. vide
*/
static ObjNode LndRoot;

/* Path jusqu'au proc qui tourne le serveur
   ****************************************
*/
static char machine_name[IOCDataMax];

/* Proto des fonction activees par le dispatcher
   *********************************************
*/
void      do_open(ServInfo * request);
void      do_delete(ServInfo * request);

/* Autres
   *******
*/
void      Acquitter(MCB * m);
void      CallLindaFunction(VoidFnPtr p, MsgBuf * msg);
int       Signal_request(MCB * mcb, int function_code);
void      RemoveTuple(TupleList * list, word num_tuple);
TupleHeader *FindTuple(TupleList *list,
			char *id,
			SELECTOR *selector) ;

/* Donnees pour le dispatcher standard d'helios
   *******************************************
*/
static DispatchInfo LndInfo =
{
	(DirNode *) & LndRoot,	/* root directory */
	NullPort,		/* port reception requetes (initialise + tard)*/
	SS_Unknown,		/* ID du sous-systeme (pour messages d'err) */
	machine_name,		/* Nom de la racine (Proc qui tourne le serveur) */
	{NULL, 0},		/* ptr fct de traitement requete User: NON UTILE */
	{	
		/* on prend des fct toutes cuites d'helios si c'est possible */
		{do_open, 2000},	/* open (fabrication maison) */
		{InvalidFn, 2000},	/* create */
		{DoLocate, 2000},	/* locate */
		{DoObjInfo, 2000},	/* objectinfo */
		{InvalidFn, 2000},	/* serverinfo */
		{do_delete, 2000},	/* delete (fabrication maison) */
		{InvalidFn, 2000},	/* rename */
		{InvalidFn, 2000},	/* link */
		{InvalidFn, 2000},	/* protect */
		{InvalidFn, 2000},	/* setdate */
		{InvalidFn, 2000},	/* refine */
		{InvalidFn, 2000},	/* close obj */
		{InvalidFn, 2000},	/* revoke */
		{InvalidFn, 2000},	/* reserve */
		{InvalidFn, 2000},	/* reserve */
	}
};

/* Decode la ligne de commande lors du lancement de serveur
***********************************************************

syntaxe:
lnd [-s numero_desire -d]

Signification des flags:

-s NUM		-> donne au serveur le numero NUM (special, a
		utiliser avec parcimonie)
		
-d		-> les serveur est un serveur de DEBUG

-r		-> (random) la strategie de gestion du tuple-space ne
		garantit PAS une gestion fifo des tuples, mais les temps
		de traitement sont optimaux. Par defaut, la strategie
		EST fifo.
		
retour 1 si OK, 0 si PB
*/
static int decode_args(int argc, char **argv)
{
	int       i;

	for (i = 1; i < argc; ++i)
	{
		if (strcmp(argv[i], "-s") == 0)
		{
			++i;
			if (sscanf(argv[i], "%d", &servnum) != 1)
				return 0;
			if (servnum < 0)
				servnum = -servnum;
		}
		else if (strcmp(argv[i], "-d") == 0)
			servdbg = 1;
		else if (strcmp(argv[i], "-r") == 0)
			fifo = 0;
		else
		{
			printf("Usage : %s [-s server_number -d -r]\n", argv[0]);
			return 0;
		}
	}
	return 1;
}



/* Point d'entree : initialisation et boucle de reception
*********************************************************/

int       main(argc, argv)
int       argc;
byte     *argv[];
{
	Object   *root;
	Object   *serv;
	NameInfo  info;
	LNDSERVMAP map;

	if (!decode_args(argc, argv))
		return -1;

	if (servdbg == 0)
	{									  /* mode serveur normal */
		if (servnum >= MaxLindaServer)
		{
			printf("Valid server number range : 0..%d\n", MaxLindaServer);
			return -2;
		}

		/* On se connecte a tous les serveurs deja actifs:
		LindaConnect renvoie une liste des serveurs presents
		*/

		LindaConnect(&map);
		if (map.numserver == MaxLindaServer)	/* trop c'est trop	*/
		{
			LindaDeconnect();
			fprintf(stderr, "%s : Too many server\n", argv[0]);
			return -3;
		}

		/* On recherche le premier numero de serveur qui ne soit pas	*/
		/* deja utilise.						*/


		if (servnum == -1)		  /* choix automatique du No de serveur
				      (servnum est initialise statiquement
				      a -1)
				   */
		{
			for (servnum = 0;
				  map.lndport[servnum] && (servnum < MaxLindaServer);
				  servnum++) ;
		}

		/* on verifie que ce No n'existe pas deja */
		if (servnum >= MaxLindaServer || map.lndport[servnum])
		{
			LindaDeconnect();
			fprintf(stderr, "%s : Bad server number (%d)\n",
					  argv[0], servnum);
			return -4;
		}

		/* on calcule le nom du serveur */
		sprintf(servname, "%s%d", LSName, servnum);
		/* on reste connecte aux autres autres serveurs :
		on a besoin de causer au 1er serveur
		pour synchroniser le serveur courant */
	}
	else
	{									  /* c'est un serveur de DEBUG */
		/* NOTE: l'option -s est ignoree
			 si l'option -d est presente) */

		/* On recherche des congeneres eventuels */
		LindaConnectD(&map);
		if (map.numserver >= 1)	  /* trop c'est trop	*/
		{
			LindaDeconnectD();
			fprintf(stderr, "%s : Too many DEBUG servers\n", argv[0]);
			return -3;
		}
		servnum = 0;
		/* on calcule le nom du serveur */
		sprintf(servname, "%sD%d", LSName, servnum);
	}

	/* On initialise l'objet racine de Linda */
	/* ceci ne sert qu'au dispatcher et a faire en sorte
	   que linda ressemble a un fichier
	*/
	InitNode(&LndRoot, servname, Type_Stream,
				0, 0x43434343);
	/* 0x43 rwd */
	InitList(&LndRoot.Contents);

	/* On cree le port de reception des requetes			*/

	if ((port = NewPort()) == NullPort)
	{
		LindaDeconnect();

		fprintf(stderr, "%s : Port allocation fails\n", argv[0]);
		return -4;
	}

	/* Voila le gros morceau... On cree un objet Helios nomme	*/
	/* "/nom_du_serveur", avec tout un tas d'infos pertinantes, 	*/
	/* entre autres le port que l'on vient de creer.		*/
	/* Un Open() sur cet objet creera un stream connecte sur ce 	*/
	/* port. cf : Helios Technical Notes, Writing an Helios server	*/

	MachineName(machine_name);	  /* recherche le nom du proc
					sur lequel tourne le serveur
					*/
	root = Locate(Null(Object), machine_name);	/* On localise /	*/

	info.Flags = Flags_StripName;
	info.Matrix = DefNameMatrix; /* Bits de protection par defaut	*/
	info.LoadData = NULL;
	info.Port = port;

	/* On creer l'objet a partir de <machine_name> */

	serv = Create(root, servname, Type_Name, sizeof (NameInfo), (byte *) & info);
	Close(root);

	if (!serv)
	{
		LindaDeconnect();
		FreePort(port);
		fprintf(stderr, "%s : Cant create entry point\n", argv[0]);
		return -5;
	}

	/* remplissage des parties de LndInfo qui ne peuvent etre connues
	   qu'a l'execution
	*/
	LndInfo.ReqPort = port;

	/* Initialisation de la table de hachage.			*/

	if (InitTupleSpace(&main_tps) == 0)
	{
		LindaDeconnect();
		Delete(serv, NULL);
		FreePort(port);
		fprintf(stderr, "%s : Cant create workspace\n", argv[0]);
		return -6;
	}

	printf("\n%s\n", LINDA_ID);
	printf("Server %s/%s installed\n", machine_name, servname);


	/* Tentative vaine et pitoyable pour intercepter les signaux	*/
	/* d'interruption. Ca n'a jamais marche, pourquoi ???		*/
	/* Il FAUT que ca marche, sinon le moindre Ctrl-C declenche par	*/
	/* un utilisateur laid et sot arretera la brillante carriere du	*/
	/* serveur. Un drame de la vie somme toute anodin si le serveur	*/
	/* ne se vengeait pas en laissant l'objet cree ci-dessus 	*/
	/* derriere lui. Tout le monde sera persuade qu'il est encore 	*/
	/* la, mais on ne pourra ni lui envoyer des requetes ni prendre	*/
	/* sa place (ca sent le reset a plein nez).			*/

	signal(SIGABRT, SIG_IGN);
	signal(SIGTERM, SIG_IGN);
	signal(SIGQUIT, SIG_IGN);

	Dispatch(&LndInfo);			  /* qui fait tout .... */

	LindaDeconnect();
	Delete(serv, NULL);
	FreePort(port);

	printf("\nServer %s/%s :\nExiting on 'Delete' request\n",
			 machine_name, servname);

	return 0;
}


/* Initialisation de la table de hachage
****************************************/

word      InitTupleSpace(tps)
tpsPtr    tps;
{
	word      hsize = _hsize;	  /*Nbre de points d'entree par defaut	*/

	InitSemaphore(&tps->lock, 1);/* Inutilise en fait	*/

	tps->hash_table = (tplPtr) calloc(_hsize, sizeof (TupleList));
	if (!tps->hash_table)
		return 0;

	while (hsize--)
	{
		/* Initialisation de chaque liste d'entetes :		*/
		/* Verrou a 1, pas de tache en attente			*/

		InitSemaphore(&tps->hash_table[hsize].lock, 1);
		InitSemaphore(&tps->hash_table[hsize].waiting, 0);
		tps->hash_table[hsize].numwait = 0;
		tps->hash_table[hsize].type_table = NULL;
	}
	return 1;
}


/* Fonction de hachage : point d'entree = f( identificateur )
*************************************************************/

word      GetHashCode(id)
byte     *id;
{
	word      i, h;

	for (i = 0, h = 0; i < MAX_TPISIZE; i++)
		h = (h * 7 + 13 + 17 * id[i]) % _hsize;

	return h;
}


/* Comparaison de 2 ids : 1 = ok
********************************/

bool      MatchID(id1, id2)
byte     *id1;
byte     *id2;
{
	return (strncmp(id1, id2, MAX_TPISIZE) == 0);
}


/* Ajout un nouvel entete de tuple dans une liste
*************************************************/

tphPtr    GetNewSlot(list)
tplPtr    list;
{
	/* Si il n'y a plus de place, on realloue une page	*/

	if (list->list_size <= list->list_alloc)
	{
		tphPtr    newl;
		word      news = list->list_size + _hpage;

		newl = (tphPtr) realloc(list->list, news * sizeof (TupleHeader));
		if (!newl)
		{
			printf("Realloc error\n");
			return 0;
		}
		list->list_size = news;
		list->list = newl;
	}

	/* On retourne l'adresse de l'entete suivant le dernier */
	/* element.						*/

	return &list->list[list->list_alloc++];
}


/* Fonction d'insertion d'un mot machine
****************************************/

void      LindaPutWord(mcb)
MCB      *mcb;
{
	word      hcode;
	tphPtr    tph;
	tplPtr    list;

	/* signaler l'operation au debugger */
	Signal_request(mcb, Msg_LndPutWord & 0x000000FF);


	/* Calcul de la fonction de hachage et recuperation de la	*/
	/* liste d'entetes correspondante				*/

	hcode = GetHashCode(((tpmPtr) mcb->Data)->id);
	list = &main_tps.hash_table[hcode];

	/* On bloque l'acces a cette liste (vu qu'on va la modifier )	*/

	Wait(&list->lock);

	/* Creation d'un espace libre en fin de liste			*/

	if ((tph = GetNewSlot(list)) != NULL)
	{
		int       i;

		/* Tout ce dont on a besoin se trouve deja dans la	*/
		/* requete : id & valeur. On a juste a recopier		*/

		memcpy(tph->id, ((tpmPtr) mcb->Data)->id, MAX_TPISIZE);
		tph->size = ((tpmPtr) mcb->Data)->size;

		/* On signale une donnee courte en mettant data a 0.	*/
		/* data qui, rappelons-le, pointe vers la zone de	*/
		/* stockage (enfin normalement, mais la non...)		*/

		tph->data = NULL;

		/* Tout c'est bien passe : FnRc = Nbre d'octets ecrits */

		mcb->MsgHdr.FnRc = sizeof (word);


		/* On reactive TOUTES les taches en attente sur cette	*/
		/* liste.						*/

		if ((i = list->numwait) != 0)
		{
			list->numwait = 0;
			while (i--)
				Signal(&list->waiting);
		}
	}
	else
		/* La c'est naze */
	{
		printf("Cant create new tuple\n");
		mcb->MsgHdr.FnRc = 0;
	}

	/* On acquitte le transfert					*/
	/* On ne cherche pas a garder le routage au niveau du serveur,	*/
	/* qui en verite je vous le dis, n'en a rien a faire de savoir	*/
	/* qui au juste lui a envoyer une requete : Flags = 0		*/

	mcb->MsgHdr.Dest = mcb->MsgHdr.Reply;
	mcb->MsgHdr.Reply = NullPort;
	mcb->MsgHdr.DataSize = 0;
	mcb->MsgHdr.ContSize = 0;
	mcb->MsgHdr.Flags = 0;
	mcb->Timeout = -1;
	mcb->Control = NULL;

	PutMsg(mcb);

	Signal(&list->lock);
	return;
}


/* Le Put() normal, le vrai, le beau, celui qui traite des blocs de
** taille quelconque.
**
********************************************************************/

void      LindaPut(mcb)
MCB      *mcb;
{
	word      i;
	word      hcode;
	tphPtr    tph;
	tplPtr    list;
	Port      port;
	byte     *mem;

	/* signaler l'operation au debugger */
	Signal_request(mcb, Msg_LndPut & 0x000000FF);


	/* Le client ayant aimablement indique la taille du bloc qu'il	*/
	/* souhaiterait voir enregister, voyons si il reste assez de 	*/
	/* place.							*/


	if ((mem = malloc(((tpmPtr) mcb->Data)->size)) == NULL)
		printf("Put : malloc error\n");

	/* On alloue un port pour le reste de la conversation. En	*/
	/* effet, la transmission des donnees n'a rien a voir avec une	*/
	/* requete et ne doit donc pas transiter par le port principal	*/

	if ((port = NewPort()) == NullPort)
	{
		free(mem);
		printf("Put : NewPort error\n");
		fflush(stdout);
	}

	mcb->MsgHdr.Dest = mcb->MsgHdr.Reply;
	mcb->MsgHdr.DataSize = 0;
	mcb->MsgHdr.ContSize = 0;
	mcb->MsgHdr.Flags = 0;
	mcb->Timeout = -1;
	mcb->Control = NULL;

	hcode = GetHashCode(((tpmPtr) mcb->Data)->id);
	list = &main_tps.hash_table[hcode];

	Wait(&list->lock);
	tph = GetNewSlot(list);

	/* En cas de probleme : FnRc = -1	*/

	if (!mem || !port || !tph)
	{
		Signal(&list->lock);

		mcb->MsgHdr.Reply = NullPort;
		mcb->MsgHdr.FnRc = -1;

		PutMsg(mcb);
		return;
	}

	/* Sinon, on initialise le tuple : nom & taille			*/

	memcpy(tph->id, ((tpmPtr) mcb->Data)->id, MAX_TPISIZE);
	tph->size = ((tpmPtr) mcb->Data)->size;
	tph->data = mem;

	/* et on indique que le transfert peut commencer : FnRc = 0	*/

	mcb->MsgHdr.Reply = port;
	mcb->MsgHdr.FnRc = 0;

	PutMsg(mcb);

	mcb->MsgHdr.Dest = port;

	do
	{
		/* Buffer = directement la zone de stockage		*/

		mcb->Data = mem;
		GetMsg(mcb);
		mem += mcb->MsgHdr.DataSize;

		/* On recoit tant qu'il y a des paquets : FnRc != 0	*/
		/* cf : lndlib -> lput()				*/

	}
	while (mcb->MsgHdr.FnRc);

	/* On reveille les taches en attente */

	if ((i = list->numwait) != 0)
	{
		list->numwait = 0;
		while (i--)
			Signal(&list->waiting);
	}

	Signal(&list->lock);
	FreePort(port);				  /* a ne pas oublier */
	mess("LNDPUT: On a fini de stocker une donnee ");

	return;
}

/* ------------------------------------------------------------------
	Reception d'un tuple Type
 ******************************************************************
 */

void      LindaOut(mcb)
MCB      *mcb;
{
	word      i;
	word      hcode;
	tphPtr    tph;
	tplPtr    list;
	Port      port;
	byte     *mem;
	TUPLE_TYPE *type;

	/* signaler l'operation au debugger */
	Signal_request(mcb, Msg_LndOut & 0x000000FF);


	/* Le client ayant aimablement indique la taille du bloc qu'il	*/
	/* souhaiterait voir enregister, voyons si il reste assez de 	*/
	/* place.							*/

	if ((mem = malloc(((tpmPtr) mcb->Data)->size)) == NULL)
		printf("\nOut : malloc error (%ld bytes requested)\n",
			((tpmPtr) mcb->Data)->size );

	/* On alloue un port pour le reste de la conversation. En	*/
	/* effet, la transmission des donnees n'a rien a voir avec une	*/
	/* requete et ne doit donc pas transiter par le port principal	*/

	if ((port = NewPort()) == NullPort)
	{
		free(mem);
		printf("\nOut : NewPort error\n");
		fflush(stdout);
	}

	hcode = GetHashCode(((tpmPtr) mcb->Data)->id);
	list = &main_tps.hash_table[hcode];
	/* recuperer le global id du type */
	type = *((TUPLE_TYPE **) (mcb->Control));
	Wait(&list->lock);
	tph = GetNewSlot(list);

	mcb->MsgHdr.Dest = mcb->MsgHdr.Reply;
	mcb->MsgHdr.DataSize = 0;
	mcb->MsgHdr.ContSize = 0;
	mcb->MsgHdr.Flags = 0;
	mcb->Timeout = -1;
	mcb->Control = NULL;

	/* En cas de probleme : FnRc = -1	*/

	if (!mem || !port || !tph)
	{
		Signal(&list->lock);

		mcb->MsgHdr.Reply = NullPort;
		mcb->MsgHdr.FnRc = -1;

		PutMsg(mcb);
		return;
	}

	/* Sinon, on initialise le tuple : nom & taille	& type 		*/

	memcpy(tph->id, ((tpmPtr) mcb->Data)->id, MAX_TPISIZE);
	tph->size = ((tpmPtr) mcb->Data)->size;
	tph->data = mem;
	tph->type_handle = type;

	/* et on indique que le transfert peut commencer : FnRc = 0	*/

	mcb->MsgHdr.Reply = port;
	mcb->MsgHdr.FnRc = 0;

	PutMsg(mcb);

	mcb->MsgHdr.Dest = port;

	do
	{
		/* Buffer = directement la zone de stockage		*/

		mcb->Data = mem;
		GetMsg(mcb);
		mem += mcb->MsgHdr.DataSize;

		/* On recoit tant qu'il y a des paquets : FnRc != 0	*/
		/* cf : lndlib -> lput()				*/

	}
	while (mcb->MsgHdr.FnRc);

	/* On reveille les taches en attente */

	if ((i = list->numwait) != 0)
	{
		list->numwait = 0;
		while (i--)
			Signal(&list->waiting);
	}

	Signal(&list->lock);
	FreePort(port);				  /* a ne pas oublier */
	mess("LNDPUT: On a fini de stocker une donnee ");

	return;
}

/* Enregistre la declaration d'un type (ie: repond a la requete Msg_LndPutType)
** Construit a partir de LindaPut()
**	++Gib, 10-94
** note: doit etre blinde (de meme que LindaPut) pour gerer
**	correctement les PB d'allocation.
**	++Gib, 5-95
** Gere maintenant les creations globales (noms uniques)
** et locales.
********************************************************************/

void      LindaPutType(mcb)
MCB      *mcb;
{
	word      i;
	word      hcode;
	tplPtr    list;
	Port      port;
	byte     *mem;
	TUPLE_TYPE *tt, **type_found;
	int 	  duplicate ;

	/* signaler l'operation au debugger */
	Signal_request(mcb, Msg_LndPutType & 0x000000FF);


	/* Le client ayant aimablement indique la taille du bloc qu'il	*/
	/* souhaiterait voir enregister, voyons si il reste assez de 	*/
	/* place.							*/

	/* duplication autorisee ?? */
	duplicate = ((tpmPtr) mcb->Data)->selec_size ;
	if ((mem = malloc(((tpmPtr) mcb->Data)->size)) == NULL)
		printf("PutType : malloc error\n");

	/* On alloue un port pour le reste de la conversation. En	*/
	/* effet, la transmission des donnees n'a rien a voir avec une	*/
	/* requete et ne doit donc pas transiter par le port principal	*/

	if ((port = NewPort()) == NullPort)
	{
		free(mem);
		printf("PutType : NewPort error\n");
		fflush(stdout);
	}

	mcb->MsgHdr.Dest = mcb->MsgHdr.Reply;
	mcb->MsgHdr.DataSize = 0;
	mcb->MsgHdr.ContSize = 0;
	/* detruire chemin de retour, le reste se fait sur un autre port */
	mcb->MsgHdr.Flags = 0;
	mcb->Timeout = -1;
	mcb->Control = NULL;

	hcode = GetHashCode(((tpmPtr) mcb->Data)->id);
	list = &main_tps.hash_table[hcode];

	Wait(&list->lock);			  /* attente entree libre */
	tt = NewTupleType();

	/* En cas de probleme : FnRc = NO_CONNECT ou NO_MEMORY	*/

	if (!mem || !port || tt == NULL)
	{
		Signal(&list->lock);

		mcb->MsgHdr.Reply = NullPort;	/* dialogue fini */
		mcb->MsgHdr.FnRc = (long) ((!port) ? NO_CONNECT : NO_MEMORY);;

		PutMsg(mcb);
		return;
	}

	/* Sinon, on initialise le type */

	memcpy(tt->name, ((tpmPtr) mcb->Data)->id, MAX_TPISIZE);
	tt->fmt = mem;					  /* descripteur de format */
	tt->glob_handle = NULL;	/* inutile pour type global */
	tt->numdef = 1;		/* une seule definition si nouveau type */
	tt->next = NULL;	/* fin de liste */

	/* et on indique que le transfert peut commencer : FnRc = 0	*/

	mcb->MsgHdr.Reply = port;
	mcb->MsgHdr.FnRc = 0;

	PutMsg(mcb);

	mcb->MsgHdr.Dest = port;

	do
	{
		/* Buffer = directement la zone de stockage		*/

		mcb->Data = mem;
		GetMsg(mcb);
		mem += mcb->MsgHdr.DataSize;

		/* On recoit tant qu'il y a des paquets : FnRc != 0	*/
		/* cf : lndlib -> lput()				*/

	}
	while (mcb->MsgHdr.FnRc);

	/* on a recu le descriptif de type: ce type est il acceptable ? */
	/* rechercher dans la table des types de cette entree */

	/* ++Gib: Le hack qui suit permet de forcer la duplication
	** d'un type, meme si il existe deja un type de meme nom
	**  ==> *type_found est TJ NULL si duplicate == true 
	*/
	type_found = FindLocalTypeByName(duplicate ? "" : tt->name, 
				&(list->type_table));

	if (*type_found != NULL)
	{
		if (strcmp((*type_found)->fmt, tt->fmt) == 0)
		{								  /* meme format : presque super */
			free(tt->fmt);
			free(tt);
			tt = *type_found;
			++(tt->numdef);		  /* une definition de + pour ce type */
		}
		else
		{								  /* la ... C'est une erreur (declaration differente) */
			free(tt->fmt);
			free(tt);
			tt = TYPE_MISMATCH;
		}
	}
	else
	{									  /* sinon, le type n'existait pas : super */
		/* parser la description */
		
		if (!analyse_fmt(tt->fmt, &(tt->decl)))
		{
			/* PB: description incorrecte */
			free(tt->fmt);
			free(tt);
			tt = TYPE_BADDESC;
		}
		else
		{
			/* on accroche le nouveau type a l'arbre */
			*type_found = tt;
		}
	}

	/*
	** Il reste a conclure le protocole avec le client...
	** C'est a dire lui renvoyer le handler global du type,
	** ou un code d'erreur.
	** Le renvoi se fait par FnRc
	*/
	/* ce port a ete cree par le client */
	mcb->MsgHdr.Dest = mcb->MsgHdr.Reply;
	mcb->MsgHdr.Reply = NullPort;

	mcb->MsgHdr.FnRc = (WORD) tt;
	mcb->MsgHdr.DataSize = 0;
	mcb->MsgHdr.ContSize = 0;
	/* le dernier qui ecrit doit liberer le chemin */
	mcb->MsgHdr.Flags = 0;
	mcb->Timeout = -1;
	mcb->Control = NULL;
	mcb->Data = NULL;
	PutMsg(mcb);

	/* On reveille les taches en attente */

	if ((i = list->numwait) != 0)
	{
		list->numwait = 0;
		while (i--)
			Signal(&list->waiting);
	}

	Signal(&list->lock);
	FreePort(port);				  /* a ne pas oublier */
	mess("LNDPUTTYPE: On a fini de stocker un type ");

	/* debug ...
	dump_types() ;
	*/
	return;
}


/* Gere les requete read, get, readu, getu, rd et in (ouf !) 
** Entree:
** mcb = MCB du message arrivant du client
** rmv = 1 si requete destructive (get, getu ou in)
**     = 0 sinon.
** sendsize = 1 si requete auto-allouante (readu, getu)
**          = 0 sinon.
** Les requetes associatives (qui necessitent un echange supplementaire
** avec le client) sont mises en oeuvre si le champs selec_size du
** TupleMsg en provenance du client est non-nul.
**
*******************************************************/

void      LindaGetRead(mcb, rmv, sendsize)
MCB      *mcb;
bool      rmv;
bool      sendsize;
{
	word      hcode;
	tphPtr    tph ;
	tplPtr    list;
	TupleMsg  *pmess;
	long selector_size ;
	word status ;
	
	/* AG, 4-93 cette primitive recoit maintenant
	** un buffer de type TupleMsg dote de toutes les
	** info necessaires au debug
	*/
	SELECTOR *selec  ;
	Port ClientPort, SelectorPort ;

	/* sauvegarde donnees arrivant, 
	** car on va peut-etre reutiliser le MCB 
	*/
	pmess = (tpmPtr) mcb->Data;  
	ClientPort = mcb->MsgHdr.Reply ;
	
	/* recherche associative ? */
	if( (selector_size = pmess->selec_size) != 0L)
	{
		/* oui, allouer le bloc de selection */
		while(( selec = (SELECTOR *)malloc( selector_size)) == NULL)
			Delay( OneSec / 4) ; /* attente 250ms */
			
		/* allouer un port pour recevoir le selecteur */
		SelectorPort = NewPort() ;
		
		/* envoyer ce port au client (eventuellement NullPort 
		** si l'allocation du port a foire)
		*/
		MakeMCB( mcb, 0 , ClientPort, SelectorPort,
			NULL, 0, NULL, 0, INFINI, PRESERVE) ;
		PutMsg( mcb ) ;

		if( SelectorPort != NullPort )
		{		
			/* Lire le bloc de selection sur ce port */
			MakeMCB( mcb, 0, 
				SelectorPort, NullPort ,
				(byte *)selec , selector_size ,
				NULL, 0 , 
				INFINI, NO_PRESERVE) ;
			status = GetMsg( mcb ) ;

			FreePort( SelectorPort ) ;
		}
	}
	else
	{
		selec = NULL ;
	}
	
	hcode = GetHashCode(pmess->id);
	list = &main_tps.hash_table[hcode];
	tph = NULL;

	do
	{
		mess("LNDGET: Attente sur liste TS");
		Wait(&list->lock);
		mess("LNDGET: Fin attente sur liste TS");


		mess("LNDGET: On recherche une donnee dans TS");

		tph = FindTuple( list , pmess->id, selec );

		/* en cas d'echec, on se met en attente sur list->waiting	*/

		if (!tph)
		{
			mess("LNDGET: Donnee non trouvee, attente sur list->waiting");
			list->numwait++;
			Signal(&list->lock);
			Wait(&list->waiting);
			mess("LNDGET: Fin attente sur list->waiting");
		}
	}
	while (!tph);
	mess("LNDGET: Donnee trouvee dans TS, on l'envoie");

	if( selec != NULL ) 
	{
		free( selec ) ;
		selec = NULL ;
	}

	/* C'est trouve, on va envoyer.	*/


	mcb->MsgHdr.Dest = ClientPort;
	mcb->MsgHdr.Reply = NullPort;
	mcb->MsgHdr.ContSize = 0;
	mcb->Timeout = -1;
	mcb->Control = NULL;

	if (sendsize)
	{
		/* Premier message contenant la taille	*/

		mcb->MsgHdr.DataSize = 0;
		mcb->MsgHdr.FnRc = (tph->data ? tph->size : sizeof (word));
		mcb->MsgHdr.Flags = MsgHdr_Flags_preserve;
		mess("LNDGET: Envoie 1er message avec taille data");
		PutMsg(mcb);
		mess("LNDGET: 1er message avec taille data envoye");
	}

	if (tph->data)
	{
		/* Cas d'une donnee longue		*/

		int       ssize = tph->size;

		mcb->Data = tph->data;

		/* On envoi par paquets (cf lndlib)	*/

		while (ssize > MAX_PSIZE)
		{
			ssize -= (mcb->MsgHdr.DataSize = MAX_PSIZE);

			mcb->MsgHdr.FnRc = 1;  /* = c'est pas fini	*/

			/* Keep in touch...		*/

			mcb->MsgHdr.Flags = MsgHdr_Flags_preserve;
			mess("LNDGET: Donnee longue envoie 1 paquet");
			PutMsg(mcb);
			mess("LNDGET: Donnee longue, fin envoi paquet");
			mcb->Data += MAX_PSIZE;
		}
		mcb->MsgHdr.DataSize = ssize;
	}
	else
	{
		/* Donnee courte			*/

		mcb->MsgHdr.DataSize = sizeof (word);
		mcb->Data = (byte *) & tph->size;
	}


	/* Envoi du dernier paquet : FnRc = 0		*/

	mcb->MsgHdr.FnRc = 0;
	mcb->MsgHdr.Flags = 0;
	mess("LNDGET: Envoie dernier paquet");
	PutMsg(mcb);
	mess("LNDGET: Fin envoi dernier paquet");


	/* Lecture destructive	*/
	if (rmv)
	{
		/* On libere la memoire du tuple		*/
		RemoveTuple( list, tph - list->list) ;
	} 
	Signal(&list->lock);
	mess("LNDGET: Fin de GET , liste deverrouillee");
	return;
}

/* Fonctions Get et Read avec envoi ou non de la taille
********************************************************/

void      LindaGet(mcb)
MCB      *mcb;
{
	/* signaler l'operation au debugger */
	Signal_request(mcb, Msg_LndGet & 0x000000FF);

	LindaGetRead(mcb, 1, 0);
}

void      LindaRead(mcb)
MCB      *mcb;
{
	/* signaler l'operation au debugger */
	Signal_request(mcb, Msg_LndRead & 0x000000FF);

	LindaGetRead(mcb, 0, 0);
}

void      LindaGetU(mcb)
MCB      *mcb;
{

	/* signaler l'operation au debugger */
	Signal_request(mcb, Msg_LndGetU & 0x000000FF);


	LindaGetRead(mcb, 1, 1);
}

void      LindaReadU(mcb)
MCB      *mcb;
{
	/* signaler l'operation au debugger */
	Signal_request(mcb, Msg_LndReadU & 0x000000FF);

	LindaGetRead(mcb, 0, 1);
}


/* Meme routine que LindaGetRead() mais non bloquante
*****************************************************/

void      LindaGetReadFast(mcb, rmv, sendsize)
MCB      *mcb;
bool      rmv;
bool      sendsize;
{
	word      i;
	word      hcode;
	tphPtr    tph;
	tplPtr    list;
	tpmPtr    pmess;				  /* AG */

	pmess = (tpmPtr) mcb->Data;
	hcode = GetHashCode(pmess->id);
	list = &main_tps.hash_table[hcode];

	Wait(&list->lock);
	for (i = 0, tph = NULL; !tph && (i < list->list_alloc); i++)
	{
		if (MatchID(pmess->id, list->list[i].id))
		{
			tph = &list->list[i];
		}
	}

	mcb->MsgHdr.Dest = mcb->MsgHdr.Reply;
	mcb->MsgHdr.Reply = NullPort;
	mcb->MsgHdr.ContSize = 0;
	mcb->Timeout = -1;
	mcb->Control = NULL;

	/* Si il n'y a pas de tuple correspondant a la demande	*/
	/* on met FnRc a zero et on retourne immediatement	*/

	if (!tph)
	{
		Signal(&list->lock);
		mcb->MsgHdr.FnRc = 0;
		mcb->MsgHdr.DataSize = 0;
		mcb->MsgHdr.Flags = 0;
		PutMsg(mcb);
		return;
	}

	/* Sinon on fait comme pour LindaGetRead()		*/

	if (sendsize)
	{
		mcb->MsgHdr.DataSize = 0;
		mcb->MsgHdr.FnRc = (tph->data ? tph->size : sizeof (word));
		mcb->MsgHdr.Flags = MsgHdr_Flags_preserve;
		PutMsg(mcb);
	}
	if (tph->data)
	{
		int       ssize = tph->size;

		mcb->Data = tph->data;
		while (ssize > MAX_PSIZE)
		{
			ssize -= (mcb->MsgHdr.DataSize = MAX_PSIZE);
			mcb->MsgHdr.FnRc = 1;
			mcb->MsgHdr.Flags = MsgHdr_Flags_preserve;
			PutMsg(mcb);
			mcb->Data += MAX_PSIZE;
		}
		mcb->MsgHdr.DataSize = ssize;
	}
	else
	{
		mcb->MsgHdr.DataSize = sizeof (word);
		mcb->Data = (byte *) & tph->size;
	}

	mcb->MsgHdr.FnRc = 0;
	mcb->MsgHdr.Flags = 0;
	PutMsg(mcb);

	if (rmv)
	{
		RemoveTuple(list, i);
	}
	Signal(&list->lock);
	return;
}

/* -------------------------------------------------------------------
	Fonction qui retire un type de la table des types
	Le type n'est effectivement detruit QUE si il y a eu autant
	de destructions que de declarations de ce type.
	Dans ce cas, tous les tuples presents AYANT CE TYPE
	sont egalement detruits.
	Le suffixe "Fast" signifie que ce Get est NON BLOQUANT
	(ie: si le type n'existe pas, il n'y a pas d'attente.
	Ce point peut etre sujet a revision si cela s'avere interessant)
	Renvoie au client:
		-1 si le type a detruire n'existe pas
		(NO_SUCH_TYPE)
		le nb d'occurence du type existant encore
		APRES destruction de l'occurence courante.
**********************************************************************/

void      LindaGetTypeFast(mcb)
MCB      *mcb;
{
	word      i, status;
	word      hcode;
	TupleHeader *tph;
	TupleList *list;
	TupleMsg *pmess;
	TUPLE_TYPE *type;

	/* signaler l'operation au debugger */
	Signal_request(mcb, Msg_LndGetTypeFast & 0x000000FF);

	pmess = (tpmPtr) mcb->Data;
	hcode = GetHashCode(pmess->id);
	list = &main_tps.hash_table[hcode];

	/* verrouiller cette entree */
	Wait(&list->lock);

	/* Recup type a detruire */
	type = (TUPLE_TYPE *)(pmess->selec_size) ;

	/* verifier si le type ID correspond bien au nom passe */
	if( strcmp(type->name , pmess->id) != 0 )
	{
		printf("\nLindaGetTypeFast: Bad Type-Id\n" ) ;
		fflush(stderr ) ;
		type = NULL ;
	}

	if (type == NULL)
	{
		status = NO_SUCH_TYPE ; /* -1 en principe */
	}
	else if ((--(type->numdef)) <= 0)
	{	/* c'etait la DERNIERE occurence */
		/* retirer le type de l'arbre: */
		DeleteLocalType(type, &(list->type_table));

		/* maintenant, detruire les tuples ayant ce type */
		for (i = 0; i < list->list_alloc; ++i)
		{
			tph = &(list->list[i]);
			if (type == tph->type_handle)
			{	/* on doit detruire ce tuple */
				RemoveTuple(list, i);
			}
		}

		/* derniere occurence */
		status = 0;
	}
	else
	{
		status = type->numdef;
	}

	/* les donnees sont propres, on n'a plus besoin d'y acceder */
	Signal(&list->lock);

	/* Reponse standard:
	** ie: comme LindaGetReadXXX()
	*/
	mcb->MsgHdr.Dest = mcb->MsgHdr.Reply;
	mcb->MsgHdr.Reply = NullPort;/* pas de reponse attendue */
	mcb->MsgHdr.ContSize = 0;
	mcb->Timeout = -1;
	mcb->Control = NULL;
	mcb->MsgHdr.FnRc = 0;		  /* dernier message */
	mcb->MsgHdr.DataSize = sizeof (word); /* un mot long */
	mcb->Data = (void *) &status;/* nb d'occurence restant */
	mcb->MsgHdr.Flags = 0;		  /* scratch du chemin */
	PutMsg(mcb);

	return;
}

/* Fonctions Get et Read version non bloquantes
************************************************/

void      LindaGetFast(mcb)
MCB      *mcb;
{

	/* signaler l'operation au debugger */
	Signal_request(mcb, Msg_LndGetFast & 0x000000FF);

	LindaGetReadFast(mcb, 1, 0);
}

void      LindaReadFast(mcb)
MCB      *mcb;
{

	/* signaler l'operation au debugger */
	Signal_request(mcb, Msg_LndReadFast & 0x000000FF);

	LindaGetReadFast(mcb, 0, 0);
}

void      LindaGetFastU(mcb)
MCB      *mcb;
{
	/* signaler l'operation au debugger */
	Signal_request(mcb, Msg_LndGetFastU & 0x000000FF);

	LindaGetReadFast(mcb, 1, 1);
}

void      LindaReadFastU(mcb)
MCB      *mcb;
{
	/* signaler l'operation au debugger */
	Signal_request(mcb, Msg_LndReadFastU & 0x000000FF);

	LindaGetReadFast(mcb, 0, 1);
}

/* **************************************************************
	Cette fonction joue le meme role que do_delete( )
	mais repond a la requete privee LindaKill
	(generee par l'utilitaire skill)
	Plutot qu'a la requete std Delete
	(generee par rm du shell par exemple)
*/
void      LindaKill(MCB * mcb)
{

	/* on acquitte seulement */
	Acquitter(mcb);

	/* provoquer la sortie de Dispatch() en tuant le port */
	Delay(OneSec / 10);
	AbortPort(port, EC_Fatal + EG_Exception + EE_Abort);
}

/* -----------------------------------------------------------------------
	Modifie le mode debug du serveur
*/
void      LindaChangeDbgLevel(MCB * mcb)
{
	int       mode;

	/* le mode desire est un int */
	mode = *((int *) mcb->Data);

	if (mode >= 0 && mode <= 3)
	{
		IOdebug("\n%s: Debug level changed from %d to %d",
				  servname, dbg_level, mode);
		dbg_level = mode;
	}
	/* accuser reception */
	Acquitter(mcb);

}

/* ************************************************************************
	Celle ci ne fait manifestement pas grand chose, et ne peut etre
	pour l'instant appelee. Elle est la pour completude.
*/
void      LindaNop(MCB * mcb)
{
	/* accuser reception */
	Acquitter(mcb);

}

/* ************************************************************************
	Gere la requete de synchronisation, qui est activee par la
	fonction LindaConnect(), apres que le client se soit connecte
	aux differents serveurs. Comme les serveurs SONT des clients
	potentiels, elle permet de synchroniser les serveurs entre eux.
	Le principe est le suivant:
	On recherche dans la table des serveurs (lndport)
	celui qui possede le plus bas numero: il est alors utilise
	comme base de temps, c'est a dire qu'on lui envoie une requete
	Synchronize, traitee par la presente fonction.
	Ceci suppose que le premier serveur est le meme, durant toute
	la sequence d'initialisation des clients.
	
*/
void      LindaSynchronize(MCB * m)
{
	/* il n'y a pas de parametre a cette requete, et la valeur renvoyee
	est simplement l'heure locale */


	if (m->MsgHdr.Reply != NullPort)
	{
		m->MsgHdr.FnRc = _cputime();

		ErrorMsg(m, Err_Null);
	}

}

/* ********************************************************************
	Impression du nom de tous les tuples du TUPLE-SPACE
	
	il n'y a pas de parametre a cette requete, et la valeur renvoyee
	est simplement 1
*/


void      LindaPrint(MCB * m)
{
	long      i, j;

	/* afficher le nom du serveur... */
	printf("\n\nLinda Server: %s/%s...\n", machine_name, servname);
	/* afficher les nom de tous les tuples presents.... */
	for (i = 0; i < _hsize; ++i)
	{
		TupleList *tl = &main_tps.hash_table[i];
		
		/* pour garantir que cette entree n'est pas en cours de modif */ 
		Wait(&tl->lock) ;

		if(	tl-> numwait != 0 || 
			tl->type_table != NULL ||
			tl->list_alloc != 0)
		{
			printf("\n ! Entry #%ld :", i ) ;
		}  
		/* Afficher le nb de process en attente */
		if( tl->numwait != 0 )
		{
			printf("\n ! %ld waiting process.", tl->numwait ) ;
		}
		/* Table des types de cette entree */
		if (tl->type_table != NULL)
		{	
			printf("\n ! The following types are defined:" ) ;
			print_types(tl->type_table);
		}
		for (j = 0; j < tl->list_alloc; ++j)
		{
			tphPtr    ptuple = &tl->list[j];

			if (*(ptuple->id))
			{							  /* entree utilisee */
				char     *typed;
				TUPLE_TYPE *t = ptuple->type_handle;

				/* est ce un tuple type ? */
				typed = (t != NULL) ? t->name : "Untyped";
				if (ptuple->data != 0)
				{						  /* tuple normal */
					printf("\n[%s] <%s> (%ld bytes)",
							 ptuple->id, typed, ptuple->size);
				}
				else
				{						  /* tuple genere par putw */
					printf("\n[%s] <%s> (4 bytes) val:%ld",
							 ptuple->id, typed, ptuple->size);
				}
			}
		}
		Signal(&tl->lock) ;
	}
	fflush(stdout);
	/* NOTE: Cette fonction peut s'executer avec ou sans acquittement */
	if (m->MsgHdr.Reply != NULL)
	{
		m->MsgHdr.FnRc = 1;		  /* pourqoi pas 1 ? */
		ErrorMsg(m, Err_Null);
	}
}

/* ************************************************************************
	La fonction qui appelle les fonctions de traitement
	Et qui de-alloue le MsgBuf passe en parametre
*/
void      CallLindaFunction(VoidFnPtr p, MsgBuf * msg)
{
	/* Note: Le MsgBuf recu est alloue dynamiquement
	** par do_open() et comporte:
	**	-un MCB
	**	-un buffer data et un buffer control
	**	-pour faire le 1er echange
	**	Le MCB est initialise pour
	**	pointer sur ces 2 buffer.
	*/

	(*p) (&msg->mcb);				  /* appel fct de traitement */
	free(msg);
}

/* ************************************************************************
	La fonction qui ouvre le serveur
	Et qui recoit toutes les requetes ulterieures,
	pour un client donne
	Voir p469 de "The Helios parallel Operating System"
	ISBN 0-13-381237-5 Prentice Hall
	
*/
void      do_open(ServInfo * servinfo)
{
	MCB      *m = servinfo->m;
	MsgBuf   *r;
	ObjNode  *f;
	char     *pathname = servinfo->Pathname;
	Port      client_port;

	f = GetTarget(servinfo);
	if (f == Null(ObjNode))
	{
		ErrorMsg(m, EO_File);
		return;
	}

	if (f != &LndRoot)
	{
		ErrorMsg(m, EC_Error + EG_WrongFn + EO_Object);
		return;
	}

	r = New(MsgBuf);
	if (r == Null(MsgBuf))
	{
		ErrorMsg(m, EC_Error + EG_NoMemory + EO_Message);
		return;
	}

	/* confectionner la reponse voir p 440 pour les Flags_... */
	FormOpenReply(r, m, f, Flags_Closeable | Flags_Server, pathname);
	/* port de reponse = port dedie a un client */
	r->mcb.MsgHdr.Reply = client_port = NewPort();
	/* envoi reponse au Open */
	PutMsg(&r->mcb);
	Free(r);

	/* on deverrouille le serveur (dans son ensemble)
	   car il possede ses propres verrous en interne
	*/
	f->Account++;					  /* pour compter le nb de clients connectes */
	UnLockTarget(servinfo);
	/* maintenant on rentre dans une boucle qui va traiter
	   toutes les requete (publiques ou privees) pour ce
	   client (voir p469)
	*/
	while (1)
	{
		word      errcode;
		MsgBuf   *msg;
		MCB      *mcb;

		/* allocation d'un MsgHdr prive */
		while ((msg = New(MsgBuf)) == NULL)
			Delay(OneSec);
		/* on aura peut etre + de chance + tard ? */

		mcb = &msg->mcb;

		mcb->MsgHdr.Dest = client_port;
		mcb->Timeout = StreamTimeout;
		/* D'apres la definition du MsgBuf, les tailles actuelles
		** des buffers sont:
		**	Data : 512 bytes 
		**	Control : 64 bytes (16 words)
		*/
		mcb->Data = msg->data;
		mcb->Control = msg->control;
		mcb->MsgHdr.Flags = 0;

		errcode = GetMsg(mcb);
		
		mcb->MsgHdr.FnRc = SS_Unknown;

		if ((errcode < FG_PrivateFirst) || (errcode > FG_PrivateLast))
		{								  /* 	requete publique */
			errcode &= FG_Mask;
			if (errcode == FG_Close)
			{
				Acquitter(mcb);
				f->Account--;		  /* un client de moins */
				free(msg);
				break;				  /* on sort sans tuer le serveur */
			}
			else if (errcode == FG_Abort)
			{
				LindaKill(m);		  /* tuer serveur */
				free(msg);
				break;
			}

		}
		else if ((errcode &= 0x0000FFFF) < MaxLindaFunction)
		{
			/* ici, un point de philosophie:
			Si le client de ce serveur ne possede PAS
			de tache forkee, un simple appel a la fonction
			de traitement suffit, puisque chaque client
			ne peut emmettre les requete QUE de maniere
			sequentielle.
			
			PAR CONTRE, si on admet qu'un client est multi-taches
			il est necessaire de traiter les requetes en //
			au niveau du serveur, donc de forker les taches
			de traitement. De la meme maniere, chaque tache
			forkee doit avoir son msgbuf prive. (voir + haut )
			*/

			if (LindaFunction[errcode] == &LindaKill)
			{							  /* pas de Fork pour cette fonction */
				LindaKill(mcb);
				free(msg);
				break;
			}
			else if (Fork(4000, (VoidFnPtr) CallLindaFunction,
							  sizeof (msg) + sizeof (*LindaFunction),
							  LindaFunction[errcode], msg) == 0)
			{
				fprintf(stderr,
					"\n%s/%s:\nInsufficient stack size.\n",
					machine_name, servname);
				fflush(stderr);
			}
			/* NOTE: C'est CallLindaFunction qui a la charge de
			deallouer le MsgBuf msg
			*/
		}
		else
		{
			SCREEN;
			fprintf(stderr, "\n%s/%s:\nUnknown private request. (%lX)\n",
					  machine_name, servname, errcode);
			fflush(stderr);
			NOSCREEN;
			Acquitter(m);
			free(msg);
		}
	}
	FreePort(client_port);

}

/* --------------------------------------------------------------------
	Acquittement general, si un port de retour est precise
*/
void      Acquitter(MCB * m)
{
	if (m->MsgHdr.Reply != NullPort)
	{
		m->MsgHdr.FnRc = 0;
		ErrorMsg(m, Err_Null);
	}
}

/* --------------------------------------------------------------------
	Traitement de la requete Delete
*/
void      do_delete(ServInfo * servinfo)
{
	MCB      *m = servinfo->m;
	ObjNode  *f;


	f = GetTarget(servinfo);
	if (f == Null(ObjNode))
	{
		ErrorMsg(m, EO_File);
		return;
	}

	if (f == (ObjNode *) & LndRoot)
	{									  /* oui, c'est bien moi */
		ErrorMsg(m, Err_Null);	  /* acquittement OK */

		/* provoquer la sortie de Dispatch() en tuant le port */

		/* si on ne laisse pas passer un peu de temps
		le AbortPort() ne fonctionne pas ... mystere */
		Delay(OneSec / 50);
		AbortPort(LndInfo.ReqPort, EC_Fatal + EG_Exception + EE_Abort);
	}
}

/* -----------------------------------------------------------------------
	Cette fonction est appelee par toutes les fonctions de
	traitement de requete qui desirent signaler la requete au debugger
	Elle se contente de recuperer les caracteristiques utiles
	de la requete et de les placer dans un tuple du "tuple-space debug"
	(DTS)
*/
int       Signal_request(MCB * mcb, int function_code)
{
	DBGREQUEST r;
	int       status;
	tpmPtr    pmess = (tpmPtr) mcb->Data; /* partie privee du message */

	/* on est en mode debug ET il existe un serveur de DTS ? */
	if (dbg_level != 0 && LindaConnectedD())
	{
		r.code_requete = function_code;	 /* code de la requete: Msg_xxxx */
		r.client_id = pmess->client_id;	 /* id du client ayant emis la requete */
		r.server_id = servnum;;	  /* c'est le No du serveur */
		r.recep_date = LindaGlobalClock(); /* heure de reception */
		r.emit_date = pmess->emit_date;	 /* heure d'emission */
		MakeTupleID(pmess->id, r.tuple_name); /* nom du tuple */
		MakeTupleID("", r.tuple_ack);	/* tuple de reponse du debugger ("" -> pas de reponse attendue) */

		status = lputD("DBG_EVNT", &r, sizeof (r));	 /* on informe le debugger */
	}
	else
		status = 0;

	return status;
}

/* ************************************************
** Routine de debug: Dump de la table des types  **
**************************************************/

void      dump_types(void)
{
	TupleList *p, *pmax;
	int       i;

	/*	Delay(OneSec) ;		*/
	SCREEN;
	for (p = main_tps.hash_table, pmax = p + _hsize, i = 0; p < pmax; ++p, ++i)
	{
		if (p->type_table != NULL)
		{
			printf("\nEntree No: %d", i);
			fflush(stdout);
			print_types(p->type_table);
		}
	}
	printf("\n -----------------------------------FIN DU DUMP\n");
	fflush(stdout);
	NOSCREEN;
}


/* ----------------------------------------------------------------------
	Cette fonction detruit LE tuple Numero k de la liste de tuples l
	Attention:
		-Ce tuple DOIT exister (pas de verif)
		-La liste des tuples est retassee selon la discipline en
		vigueur (variable fifo)
*/
void      RemoveTuple(TupleList * l, word k)
{
	TupleHeader *tph, *tphx, *tph2;

	tph = l->list + k;
	if (tph->data != NULL)
		free(tph->data);

	--l->list_alloc;		/* un element de moins */
	
	tph2 = l->list + l->list_alloc;	 /* adr dernier elem */
	if (fifo)
	{									  /* discipline fifo: on retasse betement */
		for (tphx = tph; tphx < tph2; ++tphx)
		{
			*tphx = *(tphx + 1);
		}
	}
	else
	{									  /* strategie random: on recopie le dernier element
		** sur celui qui vient d'etre detruit
		*/
		*tph = *tph2;
	}

	/* Si l'espace libre est superieur a 2 fois la page d'allocation,
	** on cherche a recuperer la place inoccupee.
	*/

	if ((l->list_size - l->list_alloc) > (_hpage * 2))
	{
		TupleHeader *tmp;
		word      nb;

		/* On ne conserve qu'une seule page de libre	*/
		nb = (l->list_alloc + _hpage) * sizeof (TupleHeader);
		tmp = (TupleHeader *) realloc(l->list, nb);
		if (tmp)
		{
			l->list_size = l->list_alloc + _hpage;
			l->list = tmp;
		}
	}
}
void      LindaIn(MCB * m)
{
	Signal_request(m, Msg_LndIn & 0x000000FF ) ;
	LindaGetRead(m, 1, 0 );
}
void      LindaRd(MCB * m)
{
	Signal_request(m, Msg_LndRd & 0x000000FF ) ;
	LindaGetRead(m, 0, 0 );
}

/* -----------------------------------------------------------------------
	Fonction de recharche d'un tuple dans une liste
	Gere la recherche par nom, ou la recherche associative
	si type_id est != de NULL
	Retour:
	L'adr du tuple, ou NULL si non trouve.
*/
TupleHeader *FindTuple(TupleList *list,		/* liste a scruter */
			char *id,		/* id classique a rechercher */
			
			/* le parametre qui suit sert 
			** a la recherche associative
			** IL DOIT ETRE NULL pour une recherche classique 
			*/
			SELECTOR *selector)	/* bloc de selection  */
{
	word 		i,j ;
	TUPLE_TYPE 	*type_id ;
	TupleHeader 	*ph ;
	char 		*real_args ;
	int 		succes ;

	if( selector == NULL )
	{	/* recherche classique */
		for( i=0, ph = list->list ; i < list->list_alloc ; ++ph, ++i)
		{
			if(MatchID(id, ph->id) ) return ph ;
		}
	}
	else
	{	/* recherche associative */
		/* quel est le type recherche ? */
		type_id = (TUPLE_TYPE *)selector->offset ;
		/* quelle est l'adr des donnees de selection */
		real_args = (char *)(selector + (selector->size+1));
/* ...		
		printf("\nOn recherche un type: %p (%s)",type_id, type_id->name) ;
		printf("\navec %d cle(s)", selector->size) ;
		fflush(stdout) ;
		for( i=1 ; i <= selector->size ; ++i)
		{
			char *qq = real_args + selector[i].arg_offset ;
			printf("\n Cle #%d Val=%c%c%c...", i , 
				qq[0], qq[1], qq[2] );
		}
*/
					
		for( i=0, ph = list->list ; i < list->list_alloc ; ++ph, ++i)
		{
			if( type_id == ph->type_handle )
			{	/* type correct */
				SELECTOR *s ;
				for( succes = j = 1 , s = selector+1 ; 
					succes && (j <= selector[0].size) ; 
						++s, ++j)
				{
					
					if( memcmp(real_args + s->arg_offset,
						ph->data + s->offset,
						s->size) != 0)
					{	/* ne correspond pas ... */
						/* optimisation */
						SELECTOR ss ;
						if( j > 1)
						{
							ss = *s ;
							*s = *(s-1) ;
							*(s-1) = ss ;
						}
						succes = 0 ;
						
					}
				}
				if( succes) return ph ;
			}
		}	/* essayer une autre entree */
	}
	return NULL ;
}		

TUPLE_TYPE *global_id(TUPLE_TYPE *t)
{
	return t->glob_handle ;
}
