
/* ---------------------------------------------------------------
**	Bibliotheque de fct d'acces aux serveurs Linda	V3.3	**
**	(C) LAMIH, equipe G.I.L	91-95				**
**	Doit etre linkee avec les programmes d'application	**
----------------------------------------------------------------*/

/* macro pour les messages de debug */
#define mess(s)
#define miss(s) \
{ fprintf(stderr,"\n(%d) %s",_cputime(), s) ; fflush(stderr ) ; }


/*		NOTE IMPORTANTE:
		***************

        Ce source est un peu special, car il permet d'obtenir deux
        bibli de fonction d'acces:

        a)      Fct d'acces aux serveurs Linda classiques, qui
        implementent une memoire repartie virtuellement partagee
        b)      Fct d'acces au serveur Linda dedie au dialogue
        entre les serveurs precedents et un debugger/profiler
        d'echange inter-process

        La seconde option est activee par la presence du symbole LNDB

        A.Gibaud, 12-92
*/
#include "linda3.h"
#include <ctype.h>

#ifndef LNDB
/* allocation exclusive fenetre pour impressions de debug */
Semaphore _screenlock = { 1, NULL, (struct Id *) & _screenlock } ;
#endif


#ifdef LNDB
#undef  MaxLindaServer
#define MaxLindaServer 1
#undef  LSName
#define LSName "lindaD"

#define LindaConnectX 	LindaConnectD
#define LindaDeconnectX LindaDeconnectD
#define SelectServerX	SelectServerD
#define LindaDbgX	LindaDbgD
#define LindaConnectedX	LindaConnectedD
#define LindaPrintTuplesX	LindaPrintTuplesD
#define lputwX		lputwD
#define lputX		lputD
#define lgetX		lgetD
#define lreadX		lreadD
#define lgetnbX		lgetnbD
#define lreadnbX	lreadnbD
#define lgetuX		lgetuD
#define lreaduX		lreaduD
#define lgetunbX	lgetunbD
#define lreadunbX	lreadunbD

#define DefineTupleTypeX	DefineTupleTypeD
#define DefineGlobalTupleTypeX	DefineGlobalTupleTypeD
#define	def_tuple_typeX		def_tuple_typeD
#define	UndefTupleTypeX		UndefTupleTypeD
#define	PutGlobalTypeX		PutGlobalTypeD
#define	outX			outD
#define CreateTupleSelectorX	CreateTupleSelectorD
#define inX			inD
#define	rdX			rdD

#else

#define LindaConnectX 	LindaConnect
#define LindaDeconnectX LindaDeconnect
#define LindaDbgX	LindaDbg
#define LindaConnectedX	LindaConnected
#define SelectServerX	SelectServer
#define LindaPrintTuplesX	LindaPrintTuples
#define lputwX		lputw
#define lputX		lput
#define lgetX		lget
#define lreadX		lread
#define lgetnbX		lgetnb
#define lreadnbX	lreadnb
#define lgetuX		lgetu
#define lreaduX		lreadu
#define lgetunbX	lgetunb
#define lreadunbX	lreadunb


#define DefineTupleTypeX	DefineTupleType
#define DefineGlobalTupleTypeX	DefineGlobalTupleType
#define	def_tuple_typeX		def_tuple_type
#define	UndefTupleTypeX		UndefTupleType
#define	PutGlobalTypeX		PutGlobalType
#define	outX			out
#define CreateTupleSelectorX	CreateTupleSelector
#define inX			in
#define	rdX			rd

static long LindaDeltaT = LONG_MAX;	/* constante a ajouter a l'heure
					locale pour avoir l'heure globale
					Tous les clients et les serveurs
					sont synchronises sur le 1er
					serveur, sauf le serveur DEBUG
					LONG_MAX => pas encore synchronise */


#endif

static long LindaClientId = -1; /* Contient l'identificateur logique
					du client (c'est en fait son numero
					de connexion). Les serveurs sont des
					clients)
					-1 => pas encore connu */



/* Prototypes des fonctions locales
***********************************
*/
static int lgetread(word req_code, byte * name,
		void **data, bool size, void *selec_bloc);

/* les types de base note: (Codable sur 3 bits) */
#define TCHAR	0
#define TBYTE	1
#define TSHORT	2
#define TLONG	3
#define TFLOAT	4
#define TDOUBLE 5
#define TBLOC	6

/* Prototypes des fonctions de gestion du typage (et acces associatif) */
int	bloc(char **p, struct decl **dd);
void	delete_decl(struct decl *d);
void	print_decl(struct decl *d);
void	print_decl1(struct decl *d);
long	count_decl(struct decl *d);
long	offset_decl(struct decl *d, long o);
int	scal(char **p, struct decl **d);
int	tableau(char **p, long *taille);
void	blks(char **) ;
long 	repetiteur( char **p) ;
DECL	*copy_decl(DECL *d) ;
DECL	**fin_seq_decl(DECL *d) ;
long	declarateur(char **p, DECL **dd) ;
int	repdecl(char **p, DECL **dd, long *taille) ;

int	selector( char **mask, DECL *tree, SELECTOR *table) ;
void	*MakeSelectBlock(SELECTOR *selec, void *data) ;
void	SelectBlockSize( SELECTOR *selec , long *header_size, long* size ) ;
struct 	xbuffer ; /* predeclaration */
int	add_xbuffer(struct xbuffer *x , char c) ;
int	expand_mask(char *mask, char **xmask) ;

char    *Strdup(char *s);
char    *Strtrim(char *s) ;
void 	*FillBuff(void *p, void *q, size_t s) ;
void	*EmptyBuff(void *p, void *q, size_t s) ;
unsigned long  IncTo( unsigned long x, unsigned long k ) ;



#ifndef LNDB
static void resynchro(void);
extern void print_types(TUPLE_TYPE *) ;
#endif

/* Tous les echanges commencent par l'emission d'une requete vers un
** serveur, et une reponse de celui-ci sur _lndreply.
** Des  ports  auxiliaires sont  ensuite  crees pour le  reste de la
** transmission. _lndlock protege l'acces au port initial.
**
*************************************************************************/

static Port _lndreply;
static Semaphore _lndlock =
{1, NULL, (struct Id *) &_lndlock};

/* Le semaphore _lndtypelock sert a proteger la table locale des types.
** (utile si le client est forme de processus legers paralleles )
**
************************************************************************/
static Semaphore _lndtypelock =
{1, NULL, (struct Id *) &_lndtypelock};


/* La localisation d'un serveur met en oeuvre un Object et un Stream
** Ces entites sont stockees dans _lndobjt et _objstrm
**
*************************************************************************/

static Object *_lndobjt[MaxLindaServer];
static Stream *_lndstrm[MaxLindaServer];


/* Les ports obtenus apres localisation sont ranges dans _lndport.
** _lndactv regroupe les ports des serveurs reellement presents,
** par exemple, si seuls les serveurs 0 et 3 sont actifs :
**
**      _lndport = [ port0, 0, 0, port3 ]
**      _lndactv = [ port0, port3, 0, 0 ] : _numserver = 2
**
*************************************************************************/


static Port _lndport[MaxLindaServer];
static Port _lndactv[MaxLindaServer];
static word _numserver;

static word _lndconfig = 0;	  /* vaut 1 si la connexion est effectuee */
static LNDSERVMAP map;			  /* map des serveurs presents */

/* copie de _lndport[], etc ... */

#ifndef LNDB
/* -----------------------------------------------------------------------
	Permet de simplifier le code:
	Remplissage d'un MCB
*/
void MakeMCB( MCB* mcb, 
		word fnrc, 
		Port dest,
		Port reply,
		byte *data, unsigned short datasize,
		word *control, unsigned char contsize,
		word timeout,
		unsigned char flags )
{
	mcb->MsgHdr.FnRc = fnrc ;
	mcb->MsgHdr.Dest = dest ;
	mcb->MsgHdr.Reply = reply ;
	mcb->MsgHdr.Flags = flags ;
	mcb->MsgHdr.DataSize = datasize ;
	mcb->MsgHdr.ContSize = contsize ;
	mcb->Timeout = timeout ;
	mcb->Control = control ;
	mcb->Data = data ;
}
#endif	
 	
/* Vrai si la procedure de connection LindaConnect() a deja ete
   effectuee avec succes
*************************************************************/

int       LindaConnectedX(void)
{
	return _lndconfig;
}

#ifndef LNDB
/* Renvoie le numero de client de l'application
*************************************************/

int       LindaClientNum(void)
{
	return LindaClientId;
}

#endif

/* Procedure de connexion a tous les serveurs actifs
******************************************************/

int       LindaConnectX(LNDSERVMAP * map)
{
	int       i;
	byte      name[16];


	if (_lndconfig)
		LindaDeconnectX();
	if ((_lndreply = NewPort()) == NullPort)
		return 0;

	for (i = _numserver = 0; i < MaxLindaServer; ++i)
	{

		/* Le nom d'un serveur est forme par "LSNameXX"     */
		/* avec XX = numero du serveur [0..MaxLindaServer[  */

		sprintf(name, "/%s%d", LSName, i);

		/* Un Locate() permet de determiner l'objet Helios  */
		/* correspondant au serveur. Un stream est cree     */
		/* vers cet objet par Open(), ce qui permet de      */
		/* recuperer le port associe a l'objet.             */
		/* NB: le numero de port obtenu n'est pas vraiment  */
		/* celui du serveur. Il s'agit en fait d'un port de */
		/* renvoi (surrogate), le routage ayant ete fait    */
		/* par le systeme au moment du Open().              */
		/* C'est l'existence de ce routage qui permettra de */
		/* dialoguer avec le serveur.                       */

		_lndport[i] = NullPort;
		if ((_lndobjt[i] = Locate(NULL, name)) != NULL)
		{
			if ((_lndstrm[i] = Open(_lndobjt[i], "", O_ReadOnly)) != NULL)
			{
				_lndport[i] = _lndstrm[i]->Server;
				_lndactv[_numserver++] = _lndport[i];
			}
			else
			{
				Close(_lndobjt[i]);
			}
		}
	}
	_lndconfig = 1;


	/*
        **	renvoyer la map des serveurs trouves:
        **	ceci permet de les connaitre sans importer
        **	de variables sensibles
        */
	for (i = 0; i < MaxLindaServer; ++i)
	{
		map->lndport[i] = _lndport[i];
		map->lndactv[i] = _lndactv[i];
	}
	map->numserver = _numserver;

	/*
	**	le code qui suit determine le numero de client
	**	(identificateur logique)
	**	de la tache en cours de connexion
	**	il n'existe pas si on se connecte au tuple space DEBUG
	*/
#ifndef LNDB
	/* tenter une connexion sur le serveur DEBUG, pour voir si il est la */
	if (!LindaConnectedD())
	{
		int       status;
		LNDSERVMAP map1;

		status = LindaConnectD(&map1);
		if (status != 0)			  /* il est la */
		{
			/* faire une lecture a la volee */
			if (lgetnbD("CLIENT_ID", &LindaClientId) <= 0)
			{
				LindaClientId = 0;
			}
			lputwD("CLIENT_ID", ++LindaClientId);
			IOdebug("New client, ID is %d", LindaClientId);
		}

	}
	/*
	**	maintenant on se synchronise, CAD que l'on calcule
	**	le decalage entre l'horloge locale et celle du premier serveur
	**	lance.
	**	NOTE: Le serveur DEBUG ne se synchronise pas, il n'est client
	**	de personne.
	*/
	LindaSynchro();				  /* 1ere synchronisation */

	/* resynchro est censee activer cycliquement Lindasynchro() */
	if (Fork(2000, resynchro, 0) == 0)
	{
		fprintf(stderr,
		"LNDLIB: Warning, can't launch synchronization process");
		fflush(stderr);
	}

#endif
	return _numserver;
}

#ifndef LNDB
/* Resynchronise periodiquement tous les clients sur le 1er serveur lance
*************************************************************************/

static void resynchro(void)
{
	/* 	si je suis le serveur de reference
	**	ou si la synchro initiale n'a pas ete faite,
	**	je ne me resynchronise pas
	**	=> sortie immediate.
	*/
	while (LindaDeltaT != 0 && LindaDeltaT != LONG_MAX)
	{
		/*
		long olddelta = LindaDeltaT ;
		*/
		Delay(OneSec * 60 * 10);  /* resynchronisation toutes les 10 min */
		LindaSynchro();
		/*
		fprintf(stderr,
		"LNDLIB: Info, client %d, Global Clock re-synchronized DeltaT: %ld->%ld .",
		LindaClientNum() , olddelta, LindaDeltaT ) ;
		fflush(stderr) ;
		*/
	}
}

#endif

/* Procedure de deconnexion
****************************/

void      LindaDeconnectX()
{
	int       i;

	for (i = 0; i < MaxLindaServer; i++)
	{
		if (_lndport[i])
		{
			Close(_lndstrm[i]);	  /* Les streams sont fermes      */
			Close(_lndobjt[i]);	  /* ainsi que les objets         */
			_lndport[i] = NullPort;
		}
	}
	_numserver = _lndconfig = 0;

	if (_lndreply)
	{
		FreePort(_lndreply);		  /* On libere le port principal  */
		_lndreply = NullPort;
	}
}

#ifndef LNDB
/* Les fonctions qui suivent n'existent que dans la version normale */

/* **********************************************************************
	Effectue la synchronisation sur le 1er serveur connu
	qui renvoie simplement son heure. Un calcul simple (voir + bas)
	permet alors de determiner le decalage entre l'horloge de reference
	et l'horloge locale. Ce calcul fait l'hypothese d'un temps de
	traitement negligeable au niveau du serveur de reference,
	ce qui est verifie par la mesure.
	( temps de traitement << a l'unite d'horloge (1/100s) )
*/
void      LindaSynchro(void)
{
	int       i;
	word      t1, t2, tremote;
	MCB       req;
	Port      synchro_port;

	/* rechercher le numero du premier serveur connu */
	for (i = 0; i < MaxLindaServer; ++i)
	{
		if (_lndport[i] != NullPort)
			break;
	}

	if (i == MaxLindaServer)
	{
		/* je suis le premier */
		LindaDeltaT = 0;
	}
	else
	{
		/*
		**	On contacte le 1er serveur,
		**	et on lui demande son heure
		**	cette requete peut arriver n'importe quand,
		**	et en particuler lorsque le port de reponse _lndreply
		**	est verrouille par une autre primitive.
		**	On doit donc utiliser ici un port de reponse prive
		*/
		if ((synchro_port = NewPort()) == NullPort)
			return;					  /* pas de port ==> pas de synchro */

		/*
		**	on n'appelle surtout pas LindaConnect(),
		**	car on n'aime pas les recursions infinies
		*/

		/* Choix du destinataire */
		req.MsgHdr.Dest = _lndport[i];

		/* indiquer au serveur ou il doit envoyer la reponse     */
		req.MsgHdr.Reply = synchro_port;


		/* Fonction =  obtenir l'heure de reference*/
		req.MsgHdr.FnRc = Msg_LndSynchronize;

		/* IMPORTANT :  le flag MsgHdr_Flags_preserve doit etre */
		/*              positionne, a defaut de quoi le systeme */
		/*              "oublierait" le routage. Il faudrait    */
		/*              refaire un Open() etc...                */

		req.MsgHdr.Flags = MsgHdr_Flags_preserve;
		req.MsgHdr.ContSize = 0;
		req.Data = NULL;
		req.MsgHdr.DataSize = 0;

		req.Timeout = -1;			  /* attente infinie      */

		/* on note l'heure locale */
		t1 = _cputime();

		PutMsg(&req);

		/* le code d'erreur recupere est l'heure du serveur
                		de reference   */

		req.MsgHdr.Dest = synchro_port;
		GetMsg(&req);

		t2 = _cputime();
		tremote = req.MsgHdr.FnRc;
		/* LindaDeltaT est calcule de facon a ce que:
                	Date_Locale + LindaDeltaT = Date_Globale
                   en supposant que la requete de synchronisation
                   met le meme temps a parvenir au serveur de reference
                   qu'a revenir
                */
		FreePort(synchro_port);
		LindaDeltaT = (tremote - ((t2 - t1) / 2)) - t1;

	}
}

/* ******************************************************************
	Renvoie l'heure globale (cad celle du serveur de reference)
	pour n'importe quel client.
	Renvoie LONG_MAX (defini dans limits.h) si la synchronisation
	n'a pas encore ete faite
*/
long      LindaGlobalClock(void)
{
	return (LindaDeltaT != LONG_MAX) ?
	_cputime() + LindaDeltaT :	  /* synchro faite */
	LONG_MAX;						  /* synchro NON faite */
}

#endif
/* Selection d'un serveur en fonction d'un identificateur
*********************************************************/

Port      SelectServerX(id)
byte     *id;
{
	if (_numserver)
	{
		word      i, h;

		/* On utilise une fonction de hachage...            */

		for (i = h = 0; i < MAX_TPISIZE; i++)
		{
			h = h * 11 + id[i] * 31 + 23;
			h %= _numserver;
		}


		/* ...et on choisi parmi les serveurs ACTIFS        */

		return _lndactv[h];
	}
	return _lndactv[0];
}


/* Construit un identificateur a partir d'une chaine de caracteres
******************************************************************/
/* commun aux version normale et debug */
#ifndef LNDB

void      MakeTupleID(name, id)
byte     *name;
byte     *id;
{
	/* Id = chaine de caracteres [+ zero de remplissage]    */

	strncpy(id, name, MAX_TPISIZE);
}

#endif

/* Envoi un mot machine vers le serveur
**
** n = lputw( nom_de_la_donnee, valeur );
** n = nombre de caracteres tranferes (normalement 4)
**     0 : pas de place sur le serveur
**    -1 : pas de serveur actif
**
***************************************/

int       lputwX(name, data)
byte     *name;
int       data;
{
	MCB       req;
	TupleMsg  msg;
	word      result;

	MakeTupleID(name, msg.id);


	/* Acces exclusif au port principal _lndreply   */

	Wait(&_lndlock);
	if (!_lndconfig && !LindaConnectX(&map))
		return -1;


	/* On choisi un serveur...                      */

	if ((req.MsgHdr.Dest = SelectServerX(msg.id)) != NullPort)
	{
		int       getmsg;

		/* indique au serveur ou il doit envoyer la reponse     */

		req.MsgHdr.Reply = _lndreply;


		/* Fonction = envoi d'une donne courte                  */

		req.MsgHdr.FnRc = Msg_LndPutWord;

		/* IMPORTANT :  le flag MsgHdr_Flags_preserve doit etre */
		/*              positionne, a defaut de quoi le systeme */
		/*              "oublierait" le routage. Il faudrait    */
		/*              refaire un Open() etc...                */

		req.MsgHdr.Flags = MsgHdr_Flags_preserve;
		req.MsgHdr.ContSize = 0;
		req.Data = (byte *) & msg;
		req.MsgHdr.DataSize = sizeof (TupleMsg);

		req.Timeout = -1;			  /* attente infinie      */
		req.Control = NULL;

		/* On envoi le message qui contient a la fois l'id et   */
		/* la valeur dans le champ size.                        */

		msg.size = data;

		/* les info de DEBUG */
		msg.client_id = LindaClientId;
		msg.emit_date = LindaGlobalClock();


		PutMsg(&req);

		/* Attend l'acquittement sur _lndreply et recupere le   */
		/* code d'erreur eventuel.                              */

		req.MsgHdr.Dest = _lndreply;
		getmsg = GetMsg(&req);

		result = req.MsgHdr.FnRc;
	}
	else
		result = -1;


	/* Re-autorise l'acces a _lndreply      */

	Signal(&_lndlock);
	return result;
}


/* Envoi un paquet de taille quelconque
**
** n = lput( nom_de_la_donnee, pointeur_sur_bloc, taille_du_bloc );
** n = nombre de caracteres tranferes (normalement taille_du_bloc)
**     0 : pas de place sur le serveur
**    -1 : pas de serveur actif
**
***************************************/

int       lputX(name, data, size)
byte     *name;
void     *data;
int       size;
{
	MCB       req;
	TupleMsg  msg;

	MakeTupleID(name, msg.id);

	mess("PUT:Attente verrou...");
	Wait(&_lndlock);
	mess("PUT:Verrou passe");

	if (!_lndconfig && !LindaConnectX(&map))
		return -1;
	mess("PUT:Lindaconnect() passe");

	/* les info de DEBUG */
	msg.client_id = LindaClientId;
	msg.emit_date = LindaGlobalClock();

	if ((req.MsgHdr.Dest = SelectServerX(msg.id)) != NullPort)
	{

		int       ssize = msg.size = size;

		mess("PUT: SelectServer() passe");

		/* Le premier message est le meme que pour putw()       */
		/* seul le code de fonction change : Msg_LndPut         */

		req.MsgHdr.Reply = _lndreply;
		req.MsgHdr.ContSize = 0;
		req.MsgHdr.DataSize = sizeof (TupleMsg);
		req.MsgHdr.FnRc = Msg_LndPut;
		req.MsgHdr.Flags = MsgHdr_Flags_preserve;
		req.Data = (byte *) & msg;
		req.Timeout = -1;
		req.Control = NULL;

		mess("PUT:Avant Envoi requete Msg_LndPut...");
		PutMsg(&req);
		mess("PUT:Apres");

		req.MsgHdr.Dest = _lndreply;


		/* Le serveur repond en indiquant si il dispose de      */
		/* la place memoire necessaire : Oui -> FnRc = 0        */
		mess("PUT: On attend la reponse du serveur...");
		GetMsg(&req);
		mess("PUT: Reponse serveur recue");


		/* Si tout est ok, le serveur a cree un nouveau port    */
		/* On peut donc liberer l'acces a _lndreply             */

		Signal(&_lndlock);
		mess("PUT:On relache le verrou");
		if (req.MsgHdr.FnRc)
		{
			mess("PUT: Le serveur dit que ca foire !");
			return 0;
		}
		else
		{
			mess("PUT: Le serveur est OK, on envoie la puree ");
		}

		/* le reste de la transmission se deroule sur le        */
		/* nouveau port.                                        */

		req.MsgHdr.Dest = req.MsgHdr.Reply;

		/* On attend plus d'acquittement...                     */

		req.MsgHdr.Reply = NullPort;


		/* Si les donnes a envoyer occupent plus de MAX_PSIZE   */
		/* le bloc est envoye par paquets. Ceci vient du fait   */
		/* que Helios ne gere pas les messages > 64k ...        */

		while (ssize > MAX_PSIZE)
		{
			ssize -= (req.MsgHdr.DataSize = MAX_PSIZE);

			/* On indique que la transmission n'est pas     */
			/* finie par FnRc = 1, et on garde les infos    */
			/* de routage                                   */

			req.MsgHdr.FnRc = 1;
			req.MsgHdr.Flags = MsgHdr_Flags_preserve;

			/* On envoie...                                  */

			req.Data = (byte *) data;
			mess("PUT: On envoie un paquet dans la boucle");
			PutMsg(&req);
			mess("PUT: Paquet envoye ");

			/* ...et on met a jour le pointeur de donnees.  */

			data = ((byte *) data) + MAX_PSIZE;
		}

		/* Dernier paquet : FnRc = 0    */

		req.MsgHdr.FnRc = 0;
		req.MsgHdr.DataSize = ssize;

		/* On peut autoriser le systeme a effacer les infos de  */
		/* routage.                                             */
		/* NB : En fait il FAUT le faire, sinon les ports de    */
		/*      renvoi situes sur la route restent alloues,     */
		/*      meme apres la fin de la transmission.           */

		req.MsgHdr.Flags = 0;

		/* C'est parti...       */

		mess("PUT: On envoie le dernier paquet");
		req.Data = (byte *) data;
		PutMsg(&req);
		mess("PUT: Dernier paquet envoye");

		return size;
	}
	return -1;
}

/* Reception...
**
** Routine privee de reception de donnees. Elle peut tout, elle connait
** tout, elle fait tout.
**
** fcode = fonction a executer, Get, Read, bloquant ou non, specif taille...
** name = je sais plus
** data = adresse d'un pointeur sur une zone memoire
** size = 1 si il faut demander la taille du bloc avant de le recevoir
**        (pour pouvoir faire la reallocation si necessaire)
**
** ++Gib: 	10-94 Cette fonction est aussi appelee pour gerer
**		la requete de destruction de type (Msg_LndGetType)
**		utilisee par lget_typenb()
**		(ce qui n'a pas necessite la moindre modif)
** ++Gib:	11-94 Cette fonction recoit maintenant 
**		un parametre supplementaire (selec_bloc)
**		qui est NON NULL si on desire faire une recherche associative
**		(ce qui complique un peu le protocole avec le serveur)
**		Voir la fonction MakeSelectBlock()
***************************************/
/* NOTE: Il existe 2 fct lgetread() 'source-identiques' :
	-Une dans la librairie normale
	-L'autre dans la libraire debug

	Elle sont toutes les 2 locales
*/
static int lgetread(fcode, name, data, size, selec_bloc)
word      fcode;
byte     *name;
void    **data;
bool      size;
void	*selec_bloc ;
{
	byte     *dt;
	int       nbytes;
	MCB       req;
	Port      reply, serv_port, SelectorPort ;
	TupleMsg  msg;
	long	selec_size, selec_header_size ;
	word status ;

	/* NOTE: (A-GIBAUD, 4-93)
		La primitive getread() utilise dorenavant un transfert par
		structure TupleMsg, contrairement aux versions 0.xx et 1.xx
		qui passaient simplement le nom du tuple a lire (char[]).
		Ceci permet de passer les informations d'estampillage
		(id logiques, heure) necessaire au debug.
	*/

	/* La reception s'effectuera sur un port different de _lndreply */
	/* pour de pas gener le multi-tache dans le cas ou la lecture   */
	/* est bloquante.                                               */

	if ((reply = NewPort()) == NullPort)
		return CONNECT_ERR;

	MakeTupleID(name, msg.id);

	mess("GET:attente de verrou");
	Wait(&_lndlock);
	mess("GET:attente verrou passee , attente connect()");

	if (!_lndconfig)
		LindaConnectX(&map);
	mess("GET:Connect() passe");

	
	if ((serv_port= SelectServerX(msg.id)) != NullPort)
	{
		/* Jusque la c'est classique.                           */
		/* Maintenant on envoi la requete de lecture : fcode    */
		/* en specifiant l'identificateur et le port de reponse */

		MakeMCB( &req, fcode, serv_port, reply, 
				(byte *)&msg , sizeof(msg),
				NULL, 0,
				INFINI , PRESERVE ) ;

		/* on estampille l'evenement 				*/
		msg.client_id = LindaClientId;
		msg.emit_date = LindaGlobalClock();

		/* EXCEPTION: Si il s'agit d'une requete de destruction de
		** type, le parametre selec_bloc n'a pas son sens habituel
		** mais doit etre passe au serveur: il s'agit de l'Id du
		** type a detruire (++Gib: 5-95) 
		*/
		if( fcode == Msg_LndGetTypeFast )
		{
			msg.selec_size = (long) selec_bloc ;
			selec_bloc = NULL ; 
		}
		else
		{
			/* recherche associative ? */
			SelectBlockSize( selec_bloc , &selec_header_size, &selec_size ) ;
			msg.selec_size = selec_size ;
		}
		/* On envoie        */
		mess("GET: On envoie requete get au serveur");
		PutMsg(&req);
		mess("GET: Requete GET envoyee");

		/* on peut deverrouiller l'acces au port */
		Signal(&_lndlock);
		mess("GET: Verrou relache");

/* ------------------------------------------------ protocole R Ass---------- */
		/* si recherche associative, 
		** on DOIT envoyer le bloc de selection 
		*/
		if( selec_bloc != NULL)
		{
			/* on lit (sur reply) le port d'envoi
			** du bloc de selection  
			** qui est alloue par le serveur 
			** et envoye dans FnRc 
			*/
			MakeMCB(&req, 0 , reply, NullPort ,
				NULL ,0 ,
				NULL ,0 ,
				INFINI, PRESERVE) ;
			GetMsg( & req ) ;
			SelectorPort = req.MsgHdr.Reply ;
			if( SelectorPort != NullPort )
			{			
				/* OK, en envoie le bloc de selection */
				MakeMCB(&req , 0 , SelectorPort , NullPort ,
					(byte *)selec_bloc , selec_size ,
					NULL, 0 ,
					INFINI , NO_PRESERVE ) ;
				status = PutMsg( &req ) ;
			}
			else
			{
				FreePort(reply) ;
				printf("\nlndlib::lgetread: Port allocation error") ;
				fflush(stdout) ;
				return CONNECT_ERR ;
			}
		 
		}			

/* ---------------- Fin de l'echange specifique a la recherche associative */

		/* On reprend (ou on continue) la conversation sur reply                 */
		/* Si on veut reallouer, il faut connaitre la taille    */
		MakeMCB( &req, 1 , reply , NullPort ,
			NULL, 0 , 	/* ne sert a rien retour par FnRc*/
			NULL, 0,
			INFINI, PRESERVE ) ;

		if (size)
		{
			/* On la recoit du serveur          */

			GetMsg(&req);

			/* Rate! par exemple : lecture non bloquante et     */
			/* donnees non presentes en memoire                 */

			if ((nbytes = req.MsgHdr.FnRc) == 0)
			{
				FreePort(reply);
				return NO_TUPLE_ERR;
			}

			/* sinon on realloue                                */

			dt = realloc(*data, (size_t) nbytes);
			if (dt == NULL)
			{
				FreePort(reply);
#ifdef LNDB
				printf("\nlndlibD::lgetread: realloc error\n");
#else
				printf("\nlndlib::lgetread: realloc error\n");
#endif
				return -1;
			}
			*data = dt;
		}

		nbytes = 0;	/* on a encore rien recu        */
		dt = (byte *) (*data);

		do		/* on recoit des paquets tant   */
		{		/* FnRc != 0 (cf : lput())      */
			req.Data = dt;
			mess("GET: On va lire un paquet");
			GetMsg(&req);
			mess("GET: On a lu un paquet");

			nbytes += req.MsgHdr.DataSize;
			dt += req.MsgHdr.DataSize;
		}
		while (req.MsgHdr.FnRc);
	}
	else
	{
		Signal(&_lndlock);	/* pas de serveur, on s'est     */
		nbytes = CONNECT_ERR ;		/* gauffre                      */
	}

	FreePort(reply);
	return nbytes;
}


/* Lit et detruit un bloc (version bloquante)
**
** n = lget( nom_de_la_donnee, pointeur_sur_bloc );
** n = nombre de caracteres tranferes (normalement taille_du_bloc)
**     0 : pas de place sur le serveur
**    -1 : pas de serveur actif
**
***************************************/

int       lgetX(name, data)
byte     *name;
void     *data;
{
	return lgetread(Msg_LndGet, name, &data, 0, NULL);
}

/* Lit sans detruire (version bloquante)
**
** n = lread( nom_de_la_donnee, pointeur_sur_bloc );
** n = nombre de caracteres tranferes (normalement taille_du_bloc)
**     0 : pas de place sur le serveur
**    -1 : pas de serveur actif
**
***************************************/

int       lreadX(name, data)
byte     *name;
void     *data;
{
	return lgetread(Msg_LndRead, name, &data, 0, NULL);
}

/* Lit et detruit (version non bloquante)
**
** n = lgetnb( nom_de_la_donnee, pointeur_sur_bloc );
** n = nombre de caracteres tranferes (normalement taille_du_bloc)
**     0 : pas de place sur le serveur
**    -1 : pas de serveur actif
**
***************************************/

int       lgetnbX(name, data)
byte     *name;
void     *data;
{
	return lgetread(Msg_LndGetFast, name, &data, 0, NULL);
}

/* Lit sans detruire (version bloquante)
**
** n = lreadnb( nom_de_la_donnee, pointeur_sur_bloc );
** n = nombre de caracteres tranferes (normalement taille_du_bloc)
**     0 : pas de place sur le serveur
**    -1 : pas de serveur actif
**
***************************************/

int       lreadnbX(name, data)
byte     *name;
void     *data;
{
	return lgetread(Msg_LndReadFast, name, &data, 0, NULL);
}

/* Lit et detruit (version bloquante avec reallocation)
**
** n = lgetu( nom_de_la_donnee, adresse_pointeur_sur_bloc );
** n = nombre de caracteres tranferes (normalement taille_du_bloc)
**     0 : pas de place sur le serveur
**    -1 : pas de serveur actif
**
***************************************/

int       lgetuX(name, data)
byte     *name;
void    **data;
{
	return lgetread(Msg_LndGetU, name, data, 1, NULL);
}

/* Lit sans detruire (version bloquante avec reallocation)
**
** n = lreadu( nom_de_la_donnee, adresse_pointeur_sur_bloc );
** n = nombre de caracteres tranferes (normalement taille_du_bloc)
**     0 : pas de place sur le serveur
**    -1 : pas de serveur actif
**
***************************************/

int       lreaduX(name, data)
byte     *name;
void    **data;
{
	return lgetread(Msg_LndReadU, name, data, 1, NULL);
}

/* Lit et detruit (version non bloquante avec reallocation)
**
** n = lgetunb( nom_de_la_donnee, adresse_pointeur_sur_bloc );
** n = nombre de caracteres tranferes (normalement taille_du_bloc)
**     0 : pas de place sur le serveur
**    -1 : pas de serveur actif
**
***************************************/

int       lgetunbX(name, data)
byte     *name;
void    **data;
{
	return lgetread(Msg_LndGetFastU, name, data, 1, NULL);
}

/* Lit sans detruire (version non bloquante avec reallocation)
**
** n = lreadunb( nom_de_la_donnee, adresse_pointeur_sur_bloc );
** n = nombre de caracteres tranferes (normalement taille_du_bloc)
**     0 : pas de place sur le serveur
**    -1 : pas de serveur actif
**
***************************************/

int       lreadunbX(name, data)
byte     *name;
void    **data;
{
	return lgetread(Msg_LndReadFastU, name, data, 1, NULL );
}

/* **********************************************************************
**	Modification de mode DEBUG du serveur
**
** status = LindaDbg( mode, num_server ) ;
** mode = niveau de debug
**     0 : pas de mode debug
**     1, 2, ou 3 : mode debug, identiques pour l'instant
** num_server = numero du serveur a titiller, ou ALLSERVERS
**
** Retour: status = -1 si aucun serveur n'a pu etre contacte
***************************************/

int       LindaDbgX(int mode, int num)
{
	MCB       req;
	int       result;
	int       how_much, i, mini, maxi;

	if (mode < 0 || mode > 3)
		return -1;

	/* Acces exclusif au port principal _lndreply   */

	Wait(&_lndlock);
	if (!_lndconfig && !LindaConnectX(&map))
		return -1;

	if (num == ALLSERVERS)
		mini = 0, maxi = MaxLindaServer - 1;
	else
		mini = maxi = num;

	for (i = mini, how_much = 0; i <= maxi; ++i)
	{

		/* Choix du destinataire */
		Port      p = _lndport[i];

		if (p == NullPort)
			continue;				  /* le serv n'existe PAS */
		else
			++how_much;

		req.MsgHdr.Dest = p;

		/* indiquer au serveur ou il doit envoyer la reponse     */
		req.MsgHdr.Reply = _lndreply;


		/* Fonction = modif niveau de debug                     */
		req.MsgHdr.FnRc = Msg_LndDbgLevel;

		/* IMPORTANT :  le flag MsgHdr_Flags_preserve doit etre */
		/*              positionne, a defaut de quoi le systeme */
		/*              "oublierait" le routage. Il faudrait    */
		/*              refaire un Open() etc...                */

		req.MsgHdr.Flags = MsgHdr_Flags_preserve;
		req.MsgHdr.ContSize = 0;
		req.Data = (byte *) & mode;
		req.MsgHdr.DataSize = sizeof (mode);

		req.Timeout = -1;			  /* attente infinie      */
		req.Control = NULL;

		PutMsg(&req);

		/* Attend l'acquittement sur _lndreply et recupere le   */
		/* code d'erreur (bidon pour l'instant)   		*/

		req.MsgHdr.Dest = _lndreply;
		GetMsg(&req);

		result = req.MsgHdr.FnRc;
	}
	result = how_much == 0 ? -1 : 1;


	/* Re-autorise l'acces a _lndreply      */
	Signal(&_lndlock);
	return result;
}

/* -----------------------------------------------------------------------
**	Affiche les noms de tous les tuples (provisoire, sert au debug)
** -----------------------------------------------------------------------
*/
int       LindaPrintTuplesX(int num)
{
	MCB       req;
	int       result;
	int       how_much, i, mini, maxi;

	/* Acces exclusif au port principal _lndreply   */

	Wait(&_lndlock);
	if (!_lndconfig && !LindaConnectX(&map))
		return -1;

	if (num == ALLSERVERS)
		mini = 0, maxi = MaxLindaServer - 1;
	else
		mini = maxi = num;

	result = 0;
	for (i = mini, how_much = 0; i <= maxi; ++i)
	{

		/* Choix du destinataire */
		Port      p = _lndport[i];

		if (p == NullPort)
			continue;				  /* le serv n'existe PAS */
		else
			++how_much;

		req.MsgHdr.Dest = p;

		/* indiquer au serveur ou il doit envoyer la reponse     */
		req.MsgHdr.Reply = _lndreply;


		/* Fonction = modif niveau de debug                     */
		req.MsgHdr.FnRc = Msg_LndPrintTuples;

		/* IMPORTANT :  le flag MsgHdr_Flags_preserve doit etre */
		/*              positionne, a defaut de quoi le systeme */
		/*              "oublierait" le routage. Il faudrait    */
		/*              refaire un Open() etc...                */

		req.MsgHdr.Flags = MsgHdr_Flags_preserve;
		req.MsgHdr.ContSize = 0;
		req.Data = NULL;
		req.MsgHdr.DataSize = 0;

		req.Timeout = -1;			  /* attente infinie      */
		req.Control = NULL;

		PutMsg(&req);

		/* Attend l'acquittement sur _lndreply et recupere le   */
		/* code d'erreur (bidon pour l'instant)   		*/

		req.MsgHdr.Dest = _lndreply;
		GetMsg(&req);

		result = req.MsgHdr.FnRc;
	}

	/* Re-autorise l'acces a _lndreply      */
	Signal(&_lndlock);
	return result;
}

#ifndef LNDB
/* -----------------------------------------------------------------------
	Renvoie le nom standard d'une requete en fonction de son numero
*/
char     *LindaRequestName(long fn)
{
	static char *m[] =
	{
		"BadFunctionNumber",
		"PutWord",
		"Put",
		"Get",
		"Read",
		"GetFast",
		"ReadFast",
		"GetU",
		"ReadU",
		"GetFastU",
		"ReadFastU",
		"KillServer",
		"DbgLevel",
		"DbgNop",
		"Synchronize",
		"PrintTuples",
		"PutType",
		"GetType",
		"Out",
		"In",
		"Rd",
	};
	int       maxi = sizeof (m) / sizeof (*m) - 1;

	fn &= 0x000000FF;

	return (fn < 0 || fn >= maxi) ? m[0] : m[fn + 1];
}

#endif
/* *******************************************************************
** Cette partie de la librairie est dediee a l'implementation 	    **
** de la notion de tuple type (T-Linda) et a l'acces associatif	    **
*********************************************************************/

/* Note: Cette table est statique, donc dupliquee lors de la generation
	des 2 versions de la librairie, sans collision d'identificateur
	aussi longtemps qu'elle ne devra pas etre accedee de l'EXTERIEUR
	de ce module.
	Par contre la fonction FindLocalType() existe a un seul exemplaire
	(dans la librairie normale)
*/
static TUPLE_TYPE *type_table = NULL;	 /* table de types connus en local */


#ifndef LNDB

/* ------------------------------------------------------------
** Deux interfaces differentes pour def_tuple_type
** (voir plus bas)
*/

/* Pour les types uniques */

TUPLE_TYPE *DefineTupleTypeX(char *name, char *fmt, ...)
{
	va_list ap ;
	TUPLE_TYPE *tt ;
	
	va_start(ap,fmt) ;
	tt = def_tuple_typeX(name,fmt, 1, ap) ;
	va_end(ap) ;
	
	return tt ;
}

/* version pour les types globaux */

TUPLE_TYPE *DefineGlobalTupleTypeX(char *name, char *fmt, ...)
{
	va_list ap ;
	TUPLE_TYPE *tt ;
	
	va_start(ap,fmt) ;
	tt = def_tuple_typeX(name,fmt, 1, ap) ;
	va_end(ap) ;
	
	return tt ;
}
/* -----------------------------------------------------------------------
	Cette fonction permet de definir un nouveau type:
	Elle procede a la definition globale du type, c'est a dire qu'elle
	informe le serveur concerne de la definition de ce type.
	Elle renvoie un 'handle' de type (TUPLE_TYPE *), ou NULL
	si la definition du type n'est pas possible.
	(Type deja defini differemment, ou pb d'allocation)
	++Gib: 16-5-95
	Le parametre supplementaire 'duplicate' permet de definir
	un type ordinaire (ie: nouveau) ou global (ie: mappe sur
	un eventuel type global existant)
	duplicate == 0 => creation normale
	duplicate == 1 => creation globale
*/
TUPLE_TYPE *def_tuple_typeX(char *name, char *fm, int duplicate, va_list ap )
{
	TUPLE_TYPE *loc, *glob_loc;
	TUPLE_TYPE **where;
	char *fmt, *sfm ; 
	int nb ;
	
	nb = strlen(fm)*2 ;
	if(nb < 30) nb = 30 ;
	sfm = calloc(nb, 1) ;
	vsprintf(sfm,fm,ap) ;
		
	fmt = Strtrim(sfm) ; free(sfm) ;
	
	Wait(&_lndtypelock);
	for (;;)
	{
		/* si duplicate == 1, on force la creation d'un nouveau
		** type, en recherchant un nom vide ==> *loc == NULL
		*/ 
		loc = *(where = FindLocalTypeByName
				(duplicate ? "" : name, &type_table));

		if (loc == NULL)
		{								  /* le type n'existe pas en local, mais peut etre en global  */
			glob_loc = PutGlobalTypeX(name, fmt, duplicate);
			if (glob_loc == TYPE_MISMATCH)
			{
				printf("\nLNDLIB3: Type def. conflicts with a previous one. %s: <%s>\n",
						 name, fmt);
				loc = NULL;
				break;
			}
			else if (glob_loc == TYPE_BADDESC)
			{
				printf("\nLNDLIB3: Bad Type Definition (parsing error). %s: <%s>\n",
						 name, fmt);
				loc = NULL;
				break;
			}
			else if (glob_loc == NO_MEMORY )
			{
				printf("\nLNDLIB3: Can't create global type : %s <%s> (no more memory)\n",
						 name, fmt);
				loc = NULL;
				break;
			}
			else if (glob_loc == NO_CONNECT)
			{
				printf("\nLNDLIB3: Can't create global type : %s <%s> (connexion error)\n",
						 name, fmt);
				loc = NULL;
				break;
			}
			/* existe en global, on peut creer en local */
			loc = NewTupleType() ;
			if (loc == NULL)
				break;
			/* construction de l'arbre lie a ce type */
			if (!analyse_fmt(fmt, &(loc->decl)))
			{
				free(loc);
				loc = NULL;
				break;
			}
			/* stockage du format */
			strcpy(loc->name, name);
			loc->fmt = Strdup(fmt);
			if (loc->fmt == NULL)
			{
				free(loc);
				loc = NULL;
				break;
			}

			loc->glob_handle = glob_loc;	 /* initialiser le handle global */
			loc->numdef = 1;		  /* inutile en principe au niveau local */
			loc->next = NULL;

			*where = loc;			  /* attacher le type a l'arbre */
		}
		else
		{								  /* type trouve, mais a il la meme definition ? */
			if (strcmp(loc->fmt, fmt) != 0)
				loc = NULL;
		}
		/* le type existe en local (et a meme definition),
		** donc en global aussi
		*/
		break;
	}
	/* la boucle for(;;) est juste la pour centraliser les
	** recuperation d'erreurs,
	** (de facon a garantir le relachement du verrou)
	** elle n'est de toute facon executee qu'une fois.
	*/
	free(fmt) ;
	/* debug */
/*
	printf("\n---TYPES LOCAUX---") ;
	print_types(type_table) ;
*/
	
	Signal(&_lndtypelock);		  /* relacher acces arbre des types */
	return loc;
}

/* ----------------------------------------------------------------
	Recherche un type dans la table des types (locale ou globale)
	(rech. seq. sur liste chainee, car les doublons sont autorises)
	Cette fct n'a pas besoin d'etre dupliquee
*/

TUPLE_TYPE **FindLocalTypeByName(char *name, TUPLE_TYPE ** root)
{
	TUPLE_TYPE **pi;

	for (pi = root; *pi != NULL; pi = &((*pi)->next))
	{
		if( strcmp(name, (*pi)->name) == 0) 
			break ;
	}

	return pi;
}

/* --------------------------------------------------------------------
	Destruction d'un type dans la table des types
	(locale ou globale, car elles ont la meme structure)
	Retour:
		NULL si le type n'existe pas dans la table
		Adresse du type detruit sinon.
	Remarque: La recherche dans la table des types peut sembler
	inutile, etant donne que l'on dispose DEJA de l'adr du type
	a detruire. En fait, cela permet de verifier que le type n'a
	pas DEJA ete detruit, donc retire de la liste.
*/
TUPLE_TYPE *DeleteLocalType(TUPLE_TYPE *t, TUPLE_TYPE **root)
{
	TUPLE_TYPE **k ;

	k = FindLocalType( t, root) ;
	if( *k != NULL )
	{
		*k = t->next ;
		free(t) ;
		return t ;
	}
	return *k ;
}
/* ------------------------------------------------------------------
** Recherche une type t dans la table, et renvoie l'adresse
** du pointeur de chainage vers ce type.
*/
TUPLE_TYPE **FindLocalType( TUPLE_TYPE *t, TUPLE_TYPE **root)
{
	for( ; *root != NULL ; root = &((*root)->next) )
	{
		if( *root == t) break ; /* trouve */
	}
	return root ;
}

/* -----------------------------------------------------------------
** Alloue et initialise un type
*/
TUPLE_TYPE *NewTupleType(void) 
{
	/* calloc() garantit une initialisation a zero */
	TUPLE_TYPE *p = calloc((size_t) 1, sizeof(TUPLE_TYPE) ) ;
	if( p == NULL )
	{
		fprintf(stderr,"\nNewTupleType: No More memory\n") ;
		fflush(stderr) ;
	}
	return p ;
}

#endif

#ifndef LNDB
/*	Fonctions de service:	
********************************************************************/

/* ---------------------------------------------------
** Copie de chaine avec alloc dynamique.
*/
char     *Strdup(char *s)
{
	size_t    t;
	char     *p = (char *) malloc(t = strlen(s) + 1);

	if (p == NULL)
		return NULL;

	memcpy((void *) p, (void *) s, t);
}

/* ------------------------------------------------------------------
** Cree une copie d'une chaine, debarassee de ses blancs
*/

char *Strtrim( char *fmt)
{
	char *f , *g, *h ;
	f = malloc( 1+strlen(fmt) );
	if( f == NULL)
	{
		fprintf(stderr, "\nStrtrim: no more memory") ;
		fflush(stdout) ;
	}
	else
	{
		for(g = f, h= fmt ; (*g = *h) != '\0' ; ++g, ++h )
		{
			if( *h == ' ')
				--g ;
		}
	}
	return f ;
}

#endif

/* ---------------------------------------------------------------------
	Creation d'un type Global.
	Cette fonction necessite un dialogue avec le serveur
	Elle est derivee de lput et peut donc necessiter une
	remise a jour en cas de modification de cette derniere.
	Retour:
		NO_MEMORY 	= PB d'allocation au niveau du serveur
		NO_CONNECT	= connexion impossible
		TYPE_MISMATCH 	= Le type existe deja en global, mais n'a pas
				la meme definition.
		(cf fichier linda3.h)
		Toute autre valeur est un handle global de type..
	++Gib: 16-5-95
	Le flag duplicate sert a indiquer si on fait une creation de type
	normale (duplicate==0) ou de type global (duplicate==1)
	Cette info est passee au serveur par le champ selec_size
	(inutilise pour cette requete)
*/
TUPLE_TYPE *PutGlobalTypeX(char *name, char *fmt, int duplicate)
{
	WORD      stat;
	MCB       req;
	TupleMsg  msg;
	Port      port;
	TUPLE_TYPE *t;
	void     *data = (void *) fmt;
	int       size = strlen(fmt) + 1;

	MakeTupleID(name, msg.id);
	/* 0 = nom de type unique , 1 = duplication autorisee */
	msg.selec_size = duplicate ;

	mess("PUT:Attente verrou...");
	Wait(&_lndlock);
	mess("PUT_TYPE:Verrou passe");

	if (!_lndconfig && !LindaConnectX(&map))
		return NO_CONNECT;
	mess("PUT_TYPE:Lindaconnect() passe");

	/* les info de DEBUG */
	msg.client_id = LindaClientId;
	msg.emit_date = LindaGlobalClock();

	/* on a besoin d'un port pour connaitre la reponse du serveur */
	port = NewPort();

	if ((port != NullPort) &&
		 (req.MsgHdr.Dest = SelectServerX(msg.id)) != NullPort)
	{

		int       ssize = msg.size = size;

		mess("PUT_TYPE: SelectServer() passe");

		/* Le premier message est le meme que pour putw()       */
		/* seul le code de fonction change : Msg_LndPutType         */

		req.MsgHdr.Reply = _lndreply;
		req.MsgHdr.ContSize = 0;
		req.MsgHdr.DataSize = sizeof (TupleMsg);
		req.MsgHdr.FnRc = Msg_LndPutType;
		req.MsgHdr.Flags = MsgHdr_Flags_preserve;
		req.Data = (byte *) & msg;
		req.Timeout = -1;
		req.Control = NULL;

		mess("PUT_TYPE:Avant Envoi requete Msg_LndPutType...");
		PutMsg(&req);
		mess("PUT_TYPE:Apres");

		req.MsgHdr.Dest = _lndreply;


		/* Le serveur repond en indiquant si il dispose de      */
		/* la place memoire necessaire : Oui -> FnRc = 0        */
		mess("PUT_TYPE: On attend la reponse du serveur...");
		GetMsg(&req);
		mess("PUT_TYPE: Reponse serveur recue");


		/* Si tout est ok, le serveur a cree un nouveau port    */
		/* On peut donc liberer l'acces a _lndreply             */

		Signal(&_lndlock);
		mess("PUT:On relache le verrou");

		t = (TUPLE_TYPE *) req.MsgHdr.FnRc;
		if (t == NO_CONNECT || t == NO_MEMORY)
		{								  /* pb de creation du port auxiliaire ou plus de memoire */
			mess("PUT_TYPE: Unable to create a new type.");
			FreePort(port);
			return t;
		}
		else
		{
			mess("PUT_TYPE: Le serveur est OK, on envoie la puree ");
		}

		/* le reste de la transmission se deroule sur le        */
		/* nouveau port, cree par le serveur.                                        */
		req.MsgHdr.Dest = req.MsgHdr.Reply;

		/* Si les donnes a envoyer occupent plus de MAX_PSIZE   */
		/* le bloc est envoye par paquets. Ceci vient du fait   */
		/* que Helios ne gere pas les messages > 64k ...        */
		/* les paquets sont envoyes sans acquittement SAUF le dernier */
		req.MsgHdr.Reply = NullPort;

		while (ssize > MAX_PSIZE)
		{
			ssize -= (req.MsgHdr.DataSize = MAX_PSIZE);

			/* On indique que la transmission n'est pas     */
			/* finie par FnRc = 1, et on garde les infos    */
			/* de routage                                   */

			req.MsgHdr.FnRc = 1;
			req.MsgHdr.Flags = MsgHdr_Flags_preserve;

			/* On envoie...                                  */

			req.Data = (byte *) data;
			mess("PUT_TYPE: On envoie un paquet dans la boucle");
			PutMsg(&req);
			mess("PUT_TYPE: Paquet envoye ");

			/* ...et on met a jour le pointeur de donnees.  */

			data = ((byte *) data) + MAX_PSIZE;
		}

		/* Dernier paquet : FnRc = 0    */
		req.MsgHdr.FnRc = 0;

		/* On attend un acquittement */
		req.MsgHdr.Reply = port;

		req.MsgHdr.DataSize = ssize;

		/* C'est parti...       */

		mess("PUT_TYPE: On envoie le dernier paquet");
		req.Data = (byte *) data;
		PutMsg(&req);
		mess("PUT_TYPE: Dernier paquet envoye");

		/* attendre la reponse du serveur sur Notre Port */
		req.MsgHdr.Dest = port;
		/* inutile d'apres la doc (p371 The Helios Par. O S )
                req.MsgHdr.ContSize    = 0;
                req.MsgHdr.DataSize    = 0;
                */
		req.Data = NULL;
		req.Timeout = -1;
		req.Control = NULL;

		stat = GetMsg(&req);
		FreePort(port);

		return (TUPLE_TYPE *) req.MsgHdr.FnRc;


	}
	return NO_CONNECT;
}


/* -----------------------------------------------------------------------
	Permet de detruire un type global
	Retour: Le nb d'occurence(s) de ce type existant encore
	sur le serveur. Si ce nb est nul, le serveur detruit
	egalement toutes les donnees de ce type.
	Le type est aussi detruit en local.
	La fct renvoie -2 si le type local etait 'herite', ce qui signifie
	qu'il n'etait pas comptabilise par le serveur de ce type.
	
	Note: Cette fonction utilise le protocole general lgetread()
	MODIFIE 
	(car le dernier parametre n'indique pas une recherche asoociative)

*******************************************************************/
int       UndefTupleTypeX(TUPLE_TYPE *t)
{
	word      i, *pi = &i;
	TUPLE_TYPE **tt ;

	/* attente table des types libre */
	Wait(&_lndtypelock) ;
	
	tt = FindLocalType( t, &type_table) ;
	if( *tt != NULL)
	{
		if( t->numdef == 0 )
		{ /* type 'herite' => rien a faire au niveau du serveur */
			i = HERITED_TYPE ; /* -2 en principe */
		}
		else
		{ /* demander au serveur de detruire le type et de renvoyer
	  	  ** le nb de declarations restantes 
	  	  */
			lgetread(Msg_LndGetTypeFast, t->name, (void **) &pi, 0, t->glob_handle );
		}
		/* detruire maintenant au nivau local */
		DeleteLocalType(t, &type_table) ; 
		/* version lente, mais maintenable */
		/*
		*tt = t->next ;
		free(t) ;
		Version plus rapide, mais posant probleme si on change
		la structre de la table 
		*/
	}
	else
	{
		i = NO_SUCH_TYPE ;
	}
	/* liberer table des types */
	Signal(&_lndtypelock) ;
	
	return i;
}

#ifndef LNDB
/* -----------------------------------------------------------------------
	Fonctions de gestion du typage
--------------------------------------------------------------------------
*/

/* ----------------------------------------------------------
**	Detruit un arbre de declaration
**
**	Entree:	d= adr racine de l'arbre de declaration a detruire
**	Note: Ne remet pas a NULL le ptr racine (passage par valeur)
**	ceci est a la charge de la fct appelante, si elle le desire
*/
void      delete_decl(DECL * d)
{
	DECL     *dd;

	for (; (d != NULL); d = dd)
	{
		dd = d->frere;
		delete_decl(d->fils);
		free(d);
	}
}

/* -----------------------------------------------------------
**	Imprime un arbre de declaration
**
**	Usage debug seulement
*/
void      print_decl(DECL * d)
{
	printf("\n\n --- Dump de l'arbre de declaration ---\n");
	print_decl1(d);
	printf("\n Nombre de declarateurs : %ld\n", count_decl(d));
}

void      print_decl1(DECL * d)
{
	int       i;
	static int print_decl_level = -1;
	static char *tnames[] = { "Char" , "Byte" , "Short" ,
				"Long", "Float", "Double" , 
				"Block" , "Internal error" } ;

	++print_decl_level;
	for (; d != NULL; d = d->frere)
	{
		printf("\n");
		for (i = print_decl_level; i > 0; --i)
			printf("  ");
		i = d->type ;
		if( i < 0 || i > 6) i = 7 ;
		printf("<%s> ", tnames[i] );
		if (d->nbelem > 1)
			printf("[%ld]", d->nbelem);
		printf(" (offset: %ld , taille: %ld)", d->offset, d->taille);
		if (d->taille == 0)
			printf("** ATTENTION ** Bloc de taille nulle.");
		print_decl1(d->fils);

	}
	--print_decl_level;
}

/* -------------------------------------------------------------
**	Compte le nombre de declarateurs presents dans un arbre
**
**	Note: N'est pas (encore ?) utilise
**
*/
long      count_decl(DECL * b)
{
	long      k = 0;

	for (; b != NULL; b = b->frere)
	{
		k += count_decl(b->fils) + 1;
	}
	return k;
}

/* -------------------------------------------------------------------
**	Calcule l'offset de chacun des declarateurs presents dans un arbre
**
**	Cette fonction doit etre appelee APRES la construction
**	de l'arbre et permet de faciliter la recherche associative
**	en autorisant un acces direct aux differents champs
**	constituants le tuple.
**	ATTENTION:
**	Pour les tableaux de structures (alias TBLOC)
**	les offsets correspondent a la 1ere structure du tableau
**	Entree:	b = Racine de l'arbre a traiter
**		k = Offset de cette racine/deb du tuple (0 pour 1er appel)
**	Sortie: Modification du declarateur pointe par b
**	Retour: Offset de l'element qui SUIT celui qui est pointe par k
**		DE MEME NIVEAU QUE CELUI CI
*/
long      offset_decl(DECL * b, long k)
{

	for (; b != NULL; b = b->frere)
	{
		b->offset = k;
		if (b->type == TBLOC)	  /* ca, ce n'est pas evident */
			k += (offset_decl(b->fils, k) - k) * b->nbelem;
		else
			k += b->taille;
	}
	return k;
}

/* -----------------------------------------------------------
**	Analyse un scalaire et renvoie son declarateur
**	E/S: 	p = &pointeur sur prochain caractere a analyser
**		d = &pointeur sur declarateur A CREER
**	Retour: 1 si il y a eu un declarateur scalaire
**		ie: c s l f d (b pour byte dans le futur ?)
*/
int       scal(char **p, DECL ** d)
{
	int       c ;
	DECL     *dd;

	blks(p) ;
	c = **p ;
	++(*p);
	*d = NULL;

	dd = newDECL(1);
	if (dd == NULL)
	{
		return 0;	  /* pour eviter des plantages */
	}


	switch (c)
	{
	case 'c':					/* char */
		dd->type = TCHAR ;
		dd->taille = 1 ;
		break ;
	case 'b':					/* byte*/
		dd->type = TBYTE ;
		dd->taille = 1;
		break;
	case 's':
		dd->type = TSHORT ;		  /* short */
		dd->taille = 2;
		break;
	case 'l':				 /* long */
		dd->type = TLONG ;
		dd->taille = 4 ;
		break ;
	case 'f':				/* float */
		dd->type = TFLOAT ;
		dd->taille = 4;
		break;
	case 'd':					  /* double */
		dd->type = TDOUBLE ;
		dd->taille = 8;
		break;
	default:
		--(*p);
		dd->taille = 0;
		break;
	}
	return dd->taille != 0 ? (*d = dd, 1) : (free(dd), 0);
}

/* -------------------------------------------------------------
**	Analyse une declaration de tableau,
**	ie: '[entier_constant_positif]'
**	E/S:	p = &pointeur sur prochain caractere a analyser
**		taille = &long ou est stockee l'indice analyse
**	Retour:	1 si il y a eu un indice, 0 sinon.
*/
int       tableau(char **p, long *taille)
{
	char     *q = *p;
	long      nb;

	*taille = 1;					  /* au cas ou on ferait un produit par ce resultat */

	if (*q++ != '[')
		return 0;

	/* ici, on doit avoir une cste entiere positive */
	for (nb = 0; *q >= '0' && *q <= '9'; ++q)
		nb = nb * 10 + (*q - '0');

	if (nb <= 0)
		return 0;

	if (*q++ != ']')
		return 0;

	*taille = nb;
	*p = q;
	return 1;
}
/* -------------------------------------------------------------
**	Saute les blancs, si il y en a
*/
void blks( char **p ) 
{
	char *q = *p ;
	while( *q == ' ' || *q == '\t' || *q == '\n' ) ++q ;
	*p = q ;
} 
/* -------------------------------------------------------------
**
** Cette fonction cree une copie recursive 
** d'un arbre de declaration
**
*/
DECL *copy_decl(DECL *d)
{
	DECL *d1 ;
	
	if( d == NULL) return NULL ;
	d1 = newDECL(1) ;
	*d1 = *d ;
	
	d1->fils = copy_decl(d->fils) ;
	d1->frere = copy_decl(d->frere) ;
	
	return d1 ;
}


/* --------------------------------------------------------------
**
**	Analyse un repetiteur (entier decimal positif)
** Renvoie 1 si pas de repetiteur et le nombre trouve sinon
**
*/
long repetiteur( char **p)
{
	long n ;
	int yadigit ;
	
	blks(p) ;
	for(n = yadigit = 0 ; isdigit(**p) ; ++(*p) )
	{
		yadigit = 1 ;
		n = n * 10 + (**p - '0') ;
	}
	blks(p) ;
	return (yadigit == 0) ? 1 : n ;
}
	
/* --------------------------------------------------------------
**
** Renvoie l'adresse du membre "frere" du dernier declarateur
** d'une sequence de DECL.
**
*/
DECL **fin_seq_decl(DECL *d)
{
	return (d->frere == NULL) ?
				&(d->frere) :
				fin_seq_decl(d->frere) ;
}
/* --------------------------------------------------------------
**
** Analyse un "declarateur" , c'est a dire un scalaire ou un
** bloc , suivi de specification(s) de tableau(x)
** (c'est a dire qque chose de la forme [...][...])
** Retour: 0 si pas de declarateur, taille de l'objet declare sinon
**
*/
long declarateur(char **p, DECL **dd)
{
	long indice, nbelem ;
	long taille ;

	blks(p) ;	/* eliminer blancs */
	
	if( bloc(p,dd) || scal(p,dd))
	{
		/* rechercher indices(s) */
		for(nbelem = 1 ; tableau(p,&indice) ; )
		{
			nbelem *= indice ;
		}
		taille = ((*dd)->taille *= nbelem) ;	/* taille en nb d'octets */
		(*dd)->nbelem = nbelem ;
	}
	else
	{
		taille = 0L;
	}
	
	return taille ;
}
/* -------------------------------------------------------------
**
** analyse un declarateur ou une sequence de declarateurs
** precede d'un facteur de repetition ou non. 
** Une sequence de declarateurs est de la forme (ddd).
** Ex:	3d ou bien 3(ddd) ou bien d
** Note: Les repetiteur nuls (ex: 0(ddd) ) sont autorises
** et invalident la declaration. (ex: {dd 0(ff) } est 
** equivalent a {dd} ).
**
** Retour: 1 si il y a eu un declarateur, 0 sinon
** Effets de bord:
** - fait evoluer le pntr d'analyse p 
** - renvoie l'adr du declarateur cree (dd)
** - la taille en octets du bloc declare 
** (utilise par la construction de blocs) 
** ou bien 0 si pas de declarateur
*/
int repdecl(char **p, DECL **dd, long *taille)
{
	long nbfois ;	/* facteur repetition */
	char *q = *p ;	/* pour retour arriere */
	DECL *sequence, **queue, *copie, *oldcopie  ;
	long ta ;
	
	/* preparons un echec */
	*taille = 0 ;
	*dd = NULL ;
	
	/* facteur de repetition present ? */
	nbfois = repetiteur(&q) ;	/* PEUT etre nul */
	
	if( *q == '(' )
	{
		/* on a une sequence */
		++q ;
		/* au moins UN declarateur doit etre present */
		if(( ta = declarateur(&q,&sequence)) == 0) 
			return 0 ;	/* declarateur absent ou mal fichu */
		for(	queue = &(sequence->frere) ;
				declarateur(&q,queue) ; 
				queue = &((*queue)->frere) )
		{
			ta += (*queue)->taille ; 
		}
		if( *q != ')' )
		{	/* syntaxe !!! */
			delete_decl(sequence) ;
			return 0 ;
		}
		++q ;
	}
	else
	{
		/* on n'a PAS de sequence */
		ta = declarateur(&q,&sequence) ;
		if(ta== 0) return 0 ;
	}
	/* On duplique l'arbre de la sequence trouvee,
	** de maniere a avoir nbfois cette sequence
	*/
	ta *= nbfois ; /* annule la taille si nbfois == 0 */
	
	for( oldcopie = sequence ; nbfois > 1 ; --nbfois)
	{
		/* methode: chaque copie est inseree en tete de la
		** sequence des copies deja creees
		*/
		copie = copy_decl(sequence) ;
		*fin_seq_decl(copie) = oldcopie ;	/* chainage en queue */
		oldcopie = copie ;
	}
	/* ici, oldcopie est la tete de la sequence des copies*/
	
	/* repetiteur nul ? */
	if( nbfois == 0)
	{ /* oui, ne pas tenir compte de ce qui vient d'etre analyse */
		delete_decl(oldcopie) ;
		oldcopie = NULL ;
		/* note: dans ce cas, la taille est forcement nulle */
	}
	*p = q ;		/* avancer ptr d'analyse */
	*dd = oldcopie ;
	*taille = ta ;
	
	return 1 ;
}
/* ------------------------------------------------------------------
**
** Analyse un bloc,  
** c'est a dire un ou plusieurs "repdecl" entre {}
** Retour:	1 si on a trouve un bloc valide, 0 sinon
** %%%%%%%%%%%%%%%%%%%%% Penser a gerer les pannes de memoire
**
*/
int bloc(char **p, DECL **dd)
{
	char *q = *p ;
	long taille ;
	
	DECL *bloc, **lastdecl, *d ;
	
	blks(&q) ;
	
	if( *q++ != '{' ) return 0 ;
	
	bloc = newDECL(1) ;
	
	bloc->taille = 0 ;
	bloc->type = TBLOC ;
	bloc->nbelem = 1 ;
	
	for(lastdecl = &(bloc->fils) ; repdecl(&q,&d,&taille) ; )
	{
		/* le declarateur peut etre correct, mais muni 
		-- d'un facteur de repetition nul
		--	==> taille nulle et  adresse nulle
		*/
		if( d != NULL )
		{
			bloc->taille += taille ;
			*lastdecl = d, lastdecl = fin_seq_decl(d) ;
		}
	}
	
	blks(&q) ;
	
	if(*q++ != '}')
	{	/* syntaxe !! */
		delete_decl(bloc) ;
		return 0 ;
	}
	*dd = bloc ;
	*p = q ;
	
	return 1 ;
}
	
	
/* -------------------------------------------------------------
**	Analyse globale du format
**	ie: construit l'arbre de declaration complet,
**      le membre taille de la racine (*d)->taille
**	est alors la taille totale du tuple.
**	Entree:	fmt = chaine C decrivant le tuple
**	E/S:	d = &pointeur sur racine a creer
**	Retour: 1 si format correct,
**		0 sinon (dans ce cas, *d est mis a NULL).
*/
int       analyse_fmt(char *fmt, DECL ** d)
{
	char     *p = fmt;
	DECL     *dd;
	int       stat;

	stat = bloc(&p, &dd) && *p == '\0';
	if (stat)
	{
		/* mise a jour des offsets */
		offset_decl(dd, 0);
		/* debug ... */
		/* print_decl(dd) ;  */
		*d = dd;
	}
	else
		*d = NULL;
	return stat;
}

#endif
/* ---------------------------------------------------------------
	LA VERITABLE FONCTION "OUT"
*************************************************************************/
int       outX(TUPLE_TYPE * t, void *data)
{
	MCB       req;
	TupleMsg  msg;
	TUPLE_TYPE *type_global = t->glob_handle;
	long      size = t->decl->taille;

	MakeTupleID(t->name, msg.id);
	/* Le type ID permet d'acceder au nom */

	mess("OUT:Attente verrou...");
	Wait(&_lndlock);
	mess("OUT:Verrou passe");

	if (!_lndconfig && !LindaConnectX(&map))
		return -1;
	mess("OUT:Lindaconnect() passe");

	/* les info de DEBUG */
	msg.client_id = LindaClientId;
	msg.emit_date = LindaGlobalClock();

	if ((req.MsgHdr.Dest = SelectServerX(msg.id)) != NullPort)
	{

		int       ssize = msg.size = size;

		mess("out: SelectServer() passe");

		/* Le premier message est le meme que pour putw()
		** mais utilise le champ Control
		** (qui ne sert pas dans les autres requetes)
		**
		** Le code de fonction est : Msg_LndOut
		*/

		req.MsgHdr.Reply = _lndreply;
		/* ContSize est exprime en MOTS DE 32 BITS */
		req.MsgHdr.ContSize = sizeof (TUPLE_TYPE *) / sizeof (word);
		req.MsgHdr.DataSize = sizeof (TupleMsg);
		req.MsgHdr.FnRc = Msg_LndOut;
		req.MsgHdr.Flags = MsgHdr_Flags_preserve;
		req.Data = (byte *) & msg;
		req.Timeout = -1;
		req.Control = (word *) & type_global;

		mess("OUT:Avant Envoi requete Msg_LndOut...");
		PutMsg(&req);
		mess("OUT:Apres");

		req.MsgHdr.Dest = _lndreply;


		/* Le serveur repond en indiquant si il dispose de      */
		/* la place memoire necessaire : Oui -> FnRc = 0        */
		mess("OUT: On attend la reponse du serveur...");
		GetMsg(&req);
		mess("OUT: Reponse serveur recue");


		/* Si tout est ok, le serveur a cree un nouveau port    */
		/* On peut donc liberer l'acces a _lndreply             */

		Signal(&_lndlock);
		mess("OUT:On relache le verrou");
		if (req.MsgHdr.FnRc)
		{
			mess("OUT: Le serveur dit que ca foire !");
			return 0;
		}
		else
		{
			mess("OUT: Le serveur est OK, on envoie la puree ");
		}

		/* le reste de la transmission se deroule sur le        */
		/* nouveau port.                                        */

		req.MsgHdr.Dest = req.MsgHdr.Reply;

		/* On attend plus d'acquittement...                     */

		req.MsgHdr.Reply = NullPort;


		/* Si les donnes a envoyer occupent plus de MAX_PSIZE   */
		/* le bloc est envoye par paquets. Ceci vient du fait   */
		/* que Helios ne gere pas les messages > 64k ...        */

		while (ssize > MAX_PSIZE)
		{
			ssize -= (req.MsgHdr.DataSize = MAX_PSIZE);

			/* On indique que la transmission n'est pas     */
			/* finie par FnRc = 1, et on garde les infos    */
			/* de routage                                   */

			req.MsgHdr.FnRc = 1;
			req.MsgHdr.Flags = MsgHdr_Flags_preserve;

			/* On envoie...                                  */

			req.Data = (byte *) data;
			mess("OUT: On envoie un paquet dans la boucle");
			PutMsg(&req);
			mess("OUT: Paquet envoye ");

			/* ...et on met a jour le pointeur de donnees.  */

			data = ((byte *) data) + MAX_PSIZE;
		}

		/* Dernier paquet : FnRc = 0    */

		req.MsgHdr.FnRc = 0;
		req.MsgHdr.DataSize = ssize;

		/* On peut autoriser le systeme a effacer les infos de  */
		/* routage.                                             */
		/* NB : En fait il FAUT le faire, sinon les ports de    */
		/*      renvoi situes sur la route restent alloues,     */
		/*      meme apres la fin de la transmission.           */

		req.MsgHdr.Flags = 0;

		/* C'est parti...       */

		mess("OUT: On envoie le dernier paquet");
		req.Data = (byte *) data;
		PutMsg(&req);
		mess("OUT: Dernier paquet envoye");

		return size;
	}
	return -1;
}

#ifndef LNDB
/* -----------------------------------------------------------------
** Expansion d'un masque de selection.
** Elle est faite par un procede de feignant, au niveau textuel.
** On cree un buffer dynamique par substitutions successives.
** Jusqu'a ce qu'il n'y ait plus de substitution possible
*/
/* Buffer dynamique: */
typedef
struct xbuffer
{
	long size ;
	long free ;
	char *base ;
} xbuffer ;

/* -------------------------------------------------------------------
** Ajout d'un caractere dans un xbuffer, qui est etendu
** dynamiquement par pas de 100 caracteres 
** Renvoie la caractere ajoute( considere commme un unsigned), 
** ou -1 en cas de PB
*/
int  add_xbuffer(xbuffer *x , char c)
{
	if( x->free >= x->size )
	{
		char *p  ; 
		p = (x->base == NULL) ? /* pour les realloc non ansi */
		  malloc(x->size + 100) :
		  realloc(x->base, x->size + 100);
		
		if( p == NULL) return -1 ;
		
		x->size += 100 ; x->base = p ;
	}
	x->base[x->free++] = c ;
	return (unsigned int)c ;
}

/* -------------------------------------------------------------------
** Extension du masque de selection
** Cette fct renvoie le nb d'expensions realisees, et se rappelle
** si il y a eu expension. Au niveau le + haut, la fct renvoie
** donc toujours 0.
** Le pointeur retourne correspond a un bloc qui DOIT ETRE
** DESALLOUE apres usage par l'appelant
*/
int expand_mask(char *mask, char **xmask)
{
	int nbx = 0, krep, out  ;
	char *m ;
	xbuffer buf ;

	buf.free = buf.size = 0L ;
	buf.base = NULL ;
	
	for( ;  *mask ; ++mask )
	{
		krep = repetiteur(&mask) ;
		/* gestion des repetiteurs nuls:
		-- on les transforme en repetiteurs 2, mais
		-- on INVALIDE l'expension , ce qui permet de "sauter"
		-- le code controle par le repetiteur.
		*/
		out = 1 ;
		if(krep == 0)
		{
			krep = 2 ; out = 0 ;
		}
		if( krep  > 1)
		{ /* on passe ici AUSSI si le repetiteur est nul */
			if (out) ++nbx ;	/* substitution en vue */
			if( *mask == '(' )
			{ /* repetition de sequence */
				while( krep-- > 0)
				{
					int cnt = 0, car ;
					m = mask ;
					/* note: l'evaluation "progressive"
					-- du && permet de ne PAS appeler
					-- add_xbuffer() si out est FAUX
					*/
					do
					{
						car = *m ;
						if( out && (add_xbuffer(&buf, car ) == -1) )
						{
							*xmask = buf.base ; return 0 ;
						}
						if( car == ')' ) --cnt ;
						else if( car == '(' ) ++cnt ;
						++m ;

					} while(*m && cnt) ;
				}
				mask = m-1 ;
			}
			else if( *mask == '{' )
			{ /* repetition de bloc */
				while( krep-- > 0)
				{
					int cnt = 0, car ;
					m = mask ;
					do
					{
						car = *m ;
						if( out && (add_xbuffer(&buf, car ) == -1) )
						{
							*xmask = buf.base ; return 0 ;
						}
						if( car == '}' ) --cnt ;
						else if( car == '{' ) ++cnt ;
						++m ;

					} while(*m && cnt) ;
				}
				mask = m-1 ;
			}
			else
			{ /* repetition d'un seul caractere */
				while( krep-- > 0)
				{
					if( out && (add_xbuffer(&buf, *mask) == -1) )
					{
						*xmask = buf.base ; return 0 ;
					}
				}
			}				
		}
		else if( add_xbuffer(&buf, *mask) == -1)
		{ /* pas de repetition */
			*xmask = buf.base ; return 0 ;
		}
	}
	
	/* ajout du nul de fin de chaine */
	if( add_xbuffer(&buf, '\0') == -1 )
	{ 
		*xmask = buf.base ; return 0 ;
	}	
	

	if( nbx == 0 ) /* pas d'expension on retourne le buffer cree*/
	{
		*xmask = buf.base ;
	}
	else
	{
		nbx = expand_mask( buf.base, xmask ) ;
		free(buf.base ) ;
	}
	return nbx ;
}
#endif

/* -----------------------------------------------------------------------
	Fabrique un selecteur associatif
	
	Un selecteur est une table t de SELECTORS organisee comme suit:
	t[0].size 	-> Nb d'entrees dans la table, sans l'entree 0
	t[0].offset 	-> type_id (converti en long) du type a acceder.
	Chacune des autres entrees correnpond a une cle d'acces:
	t[x].size	-> Taille de la cle, en bytes,
	t[x].offset	-> Position relative de la cle dans le tuple
	t[x].arg_offset	-> Position relative de l'argument reel de in() ou rd()
			a comparer a cette cle.
			
	Le selecteur est alloue dynamiquement, et il est a la charge
	du client de le detruire (avec free) lorsqu'il n'en a plus besoin.

	Arguments:
		type_id = type a rechercher
		mask	= masque de recherche.
		 	Voir la fonction selector() pour avoir le format du masque.
			
	
	Retour: 
		Pointeur sur ce selecteur, si OK
		ou NULL si
			-pas assez de memoire
			-le selecteur ne peut pas correspondre au type
			precise.



*/
SELECTOR *CreateTupleSelectorX( TUPLE_TYPE *type_id, char *mask, ... ) 
{
	int nbkey , i ;
	char *s , *xmask, *xmask1 , *smask;
	SELECTOR *sel ;
	va_list ap ;
	
	if( type_id == NULL )
		return NULL ;
		
	/* substituer le selecteur (ie: remplacer les %d par leur valeur) */
	i = strlen(mask) * 2 ;
	if( i < 30  ) i = 30 ;
	smask = calloc(i, 1) ;
	if( smask == NULL ) return NULL ;
	va_start(ap, mask) ;
	vsprintf(smask, mask, ap) ;
	va_end(ap) ;
	
	/* debug ... */
	/*	 printf("\nMasque substitue: <%s>", smask) ;  */
	
	/* expenser le selecteur */
	expand_mask(smask, &xmask) ; 
	free(smask) ;
	
	/* debug ... */
	/*	printf("\nMasque expanse: <%s>", xmask) ; */ 
			
	/* calcul de la taille de la table */
	for( nbkey = 0 , s= xmask; *s ; ++s)
		if( *s == '.' ) ++nbkey ; 
	sel = (SELECTOR *) calloc( (nbkey+1) , sizeof(*sel)) ;
	if (sel == NULL )return NULL ;
	
	/* en-tete */
	sel->offset = (long) type_id ;
	sel->size = 0 ;
	sel->arg_offset = 0 ;
	
	xmask1 = xmask ;
	if( selector( &xmask1, type_id->decl, sel ) == 0 || sel->size != nbkey )
	{
		free(sel) ; free(xmask) ;
		return NULL;
	}
	free(xmask) ;
	/* maintenant, generer les decalages des cles dans le bloc
	** des arguments reels.
	*/
	for( i = 2 , sel[1].arg_offset = 0; i <= nbkey; ++i)
		sel[i].arg_offset = sel[i-1].arg_offset + sel[i-1].size ;
	/* marquer le selecteur "OK" */
	sel->arg_offset = MAGIC_SELECTOR ;
	return sel ;
}

#ifndef LNDB

/* -----------------------------------------------------------
	Genere la table des selecteurs, pour un acces rapide
	aux tuples.
	La premiere entree de cette table est reservee comme suit:
		t[0].size 	-> nb d'entrees (SANS l'entree 0)
				Elle DOIT etre initialisee (a 0) avant le 1ere appel
				de selector().
		t[0].offset -> identificateur global du type auquel
				se rapporte cette table. (cette entree n'EST PAS
				maintenue par la fct selector() et doit etre initialisee
				autrement.		
	  Les autres entrees contiennent ce qu'il faut pour acceder aux
	  differents champs.

	Retour:
		1 = OK
		0 = Masque de selection (mask) incompatible avec le type,
			ou mal forme. Dans ce cas t[0].size est garanti etre == 0.
			Note: Ceci peut aussi arriver si le descripteur de type
			est mal foutu ce qui ne peut pas arriver en theorie.
			Le masque doit etre de la meme forme que le descripteur de type
			avec un . pour chaque parametre reel, 
			et   un ? pour chaque parametre formel.
			Ex:
			descripteur: 	{llf{dd}}
			masque:				{?.?{?.}}
			
*/
int	selector( char **mask,	DECL *tree,  SELECTOR *table)
{
	char *s = *mask ;
	DECL *pere ;

	blks( &s ) ;	
	if( *s != '{' || tree == NULL /* securite */)
	{
		return (int)(table[0].size = 0) ;
	}
	++s ;
	pere = tree ; tree = tree->fils ;
	while( *s && *s != '}')
	{
		blks( &s ) ;
		if( tree == NULL )
		{
			return (int)(table[0].size = 0) ;
		}
			
		if( *s == '{' )
		{	/* deb de bloc */
			if( ! selector( &s, tree, table ) )
			{
				return (int)(table[0].size = 0) ;
			}
		}
		else if( *s == '.')
		{	/* parametre reel => generer une entree dans la table */
			int k = ++table[0].size ;
			
			table[k].size 	= tree->taille ;
			table[k].offset = tree->offset ;
			table[k].nbelem = 	(tree->nbelem & 0x1fffffffL) | 
						((long)tree->type << 29) ;
			++s ; 
			tree = tree->frere ;
		}
		else if( *s == '?')
		{	/* parametre formel => 
			-- juste sauter cet element dans le descripteur 
			*/
			++s ;
			tree = tree->frere ;
		}
		else if( *s == '(' || *s == ')' )
		{	/* separateurs de sequences, ignorer */
			++s ;
		}
		blks( &s ) ;
	}
	blks( &s ) ;
	if( *s != '}' )
		return (int)(table[0].size = 0) ;
	tree = pere ;
	++s ;
	*mask = s ;
	return 1 ;
}
/* -------------------------------------------------------------------------
	Construit un "bloc de selection" 
	pour une requete associative particuliere in() ou rd().
	Ce bloc est compose de:
		-La table des selecteurs pour ce type de recherche associative
		(creee par lselect()
		-Les arguments reels de in() ou rd(), 
		compactes en un seul bloc.
	Il est alloue dynamiquement, et doit donc etre detruit 
	(par free() )apres usage.
	
	Arguments:
		selec = table des selecteurs,
		data  = adresse d'un bloc de donnees (typiquement, une structure)
		qui contient les parametres reels (c-a-d les cles de selection)
		ET les parametre formels, (c-a-d ceux qui seront affectes par
		in() ou rd().
	Retour:
		Adresse du bloc de selection genere ou NULL si PB.
*/
void *MakeSelectBlock(SELECTOR *selec, void *data)
{
	long size, header_size ;
	int i ;
	char *p, *bloc ;
	SELECTOR0 *s ;
	
	/* verification du magic-number de la table des selecteurs. */
	if( selec[0].arg_offset != MAGIC_SELECTOR )
		return NULL ;
		
	/* calcul de la taille de la table des selecteurs */
	SelectBlockSize( selec, &header_size, &size ) ;
		
	p = bloc = calloc( size , 1 ) ;
	if( bloc == NULL )
		return NULL ;
	
	/* copie de la table des selecteurs */
	memcpy((void *)bloc, (void *) selec, (size_t) header_size ) ;
	
	/* copie arg reels */
	for( i = 1, p += header_size ; i <= selec[0].size ; ++i)
	{
		char *source, *dest ;
		source = (((char *)data)+selec[i].offset) ;
		dest = p ;
		memcpy( (void *)dest, (void *)source , (size_t) selec[i].size) ;
		p += selec[i].size ;
	}
	/* Le type id figurant dans le selecteur fourni est un type LOCAL
	** le bloc de selection etant destine a etre envoye au serveur qui 
	** gere ce type, on doit le remplacer par son equivalent GLOBAL
	*/
	s = (SELECTOR0 *)bloc ;
	s->type_id = s -> type_id -> glob_handle ;
	
	return (void *)bloc ;
}
/* ----------------------------------------------------------------
	Calcule la taille d'un bloc de selection et de son en-tete
*/
void SelectBlockSize( SELECTOR *selec , long *header_size, long* size )
{
	int i ;
	if (selec == NULL)
	{
		*header_size = *size = 0 ;
		return ;
	}	
	*size = *header_size = (selec[0].size +1 ) * sizeof( *selec ) ;
	/* calcul de la taille totale par ajout de la taille des 
	** paralmetres reels 
	*/
	for( i = 1 ; i <= selec[0].size ; ++i)
		*size += selec[i].size ;
}
#endif

/* ----------------------------------------------------------------
	Veritable fonction in
	
	Parametres:
		selec = table de selecteurs
		data  = ptr sur bloc de donnees contenat a la fois 
			les parametres reels et formels.
			(typiquement, une structure PARTIELLEMENT remplie)
	Retour:
		taille du tuple lu. ou -1 en cas de PB.
*/
int	inX( SELECTOR *selec, void *data )
{
	char *name ;
	void *bloc ;
	int  status ;

	/* le tuple a rechercher a le nom du type
	** (pour la compatibilite avec l'acces non associatif)
	*/
	name = ((TUPLE_TYPE *)selec[0].offset)->name ;
	
	bloc = MakeSelectBlock( selec, data) ;
	if( bloc == NULL ) 
		return MEMORY_ERR  ;
	{
		SELECTOR *ps ;
		long *pl   ;
		
		ps = (SELECTOR *) bloc ;
		pl = (long *)(ps + ps->size + 1) ;
/*
		printf("\nNb de selecteurs : %ld", ps->size ) ;
		printf("\nType id global : %lX , Magic : %lX", 
				ps->offset , ps->arg_offset) ;
		for( i =1 ; i <= ps->size ; ++i, ++pl)
			printf("\n arg %ld = %ld",i,*pl) ;  
*/
	}
	/* Note: La lecture est NON reallouante */
	status = lgetread(Msg_LndIn, name , &data , 0 , bloc ) ;
	
	free(bloc) ;

	return status ;
}
/* ----------------------------------------------------------------
	Veritable fonction rd
	
	Parametres:
		selec = table de selecteurs
		data  = ptr sur bloc de donnees contenat a la fois 
			les parametres reels et formels.
			(typiquement, une structure PARTIELLEMENT remplie)
	Retour:
		taille du tuple lu. ou -1 en cas de PB.
*/
int	rdX( SELECTOR *selec, void *data )
{
	char *name ;
	void *bloc ;
	int  status ;

	/* le tuple a rechercher a le nom du type
	** (pour la compatibilite avec l'acces non associatif)
	*/
	name = ((TUPLE_TYPE *)selec[0].offset)->name ;
	
	bloc = MakeSelectBlock( selec, data) ;
	if( bloc == NULL ) 
		return MEMORY_ERR ;
	/* Note: La lecture est NON reallouante */
	status = lgetread(Msg_LndRd, name , &data , 0 , bloc ) ;
	
	free(bloc) ;

	return status ;
}
#ifndef LNDB
/* ------------------------------------------------------------------
**	Permet de copier une chaine dans un tableau de char,
**	en completant le tableau par des '\0'
*/
char *strtoarray( char *to, char *from, size_t n )
{
	size_t k ;
	char *oldto = to ;
	
	for( k=0 ; k < n && *from ; ++k, ++to, ++from)
		*to = *from ;
	for( ; k < n ; ++k,++to)
		*to = '\0' ;
	return oldto ;
}

#include <stdarg.h>
/* --------------------------------------------------------------------------
**
** La definition de va_arg adaptee a mon probleme:
** Elle n'est probablement pas portable, et sa re-ecriture
** necessite de consulter stdarg.h (ou son equivalent UNIX)
** sur la machine cible.
*/
#define my_va_arg( to , app, size ) \
{ \
	memcpy((void *)(to) , (void *) (* app ) , (size_t) size) ; \
	*app += size ; \
}
/* ------------------------------------------------------------------------
** 
** Cette fontion 'remplit' un bloc de donnees (typiquement une structure)
** en fonction de la table des selecteurs s. Elle permet de ne pas avoir
** a faire ce boulot manuellement, et de passer les cles de selection 
** comme de simples parametres de fonction.
** FillTuple() peut etre utilisee seule, et le tuple obtenu peut
** etre passe a une des fonctions de selection associative in() ou rd()
** Elle pourrait etre egalement utilisee en interne par In() et Rd(), 
** surcouches des precedentes.
**
** FillTuple est emminement NON-PORTABLE, car elle fait appel a une
** version maison de la macro ANSI va_arg, et tient compte des mecanismes
** du C ANSI lors du passage des parametres, cad:
** char et short promus en int
** float promu en double.
** Structures passees par valeur (ce n'est pas le cas en C K&R)
** Tableaux passes par adresse. (sauf si il sont membres d'une structure.)
**
** Noter egalement le traitement particulier des tableau de 'char', qui
** respecte la semantique des chaines de caracteres du C. Pour avoir
** une semantique 'tableau' ordinaire, utiliser des tableaux de 'byte'
*/
void FillTuple( void *tuple, TupleSelector s, ...)
{
	va_list ap ;
	short 	sh ;
	byte 	by ;
	int 	in ;
	char 	*p = (char *)tuple ;
	float	flo ;
	double  dou ;
	
	long nbsel , nbelem ;
	int i , type ;
	
	va_start(ap ,s ) ;
	
	nbsel = s[0].size ;
/*	printf("\n %d elements dans tab de selec", nbsel) ; */
	
	for( i = 1 ;  i <= nbsel ; ++i )
	{
		/* calculer le nb d'elements, et le type */
		nbelem = s[i].nbelem & 0x1FFFFFFFL ; /* masquer 3 bits de poid fort */
		type = s[i].nbelem >> 29 ;	/* unsigned -> pas de pb de bit de signe */
/*		printf("\n Nbelem = %d type = %d", nbelem,type) ; */
		fflush(stdout) ;
		
		if( nbelem == 1 )
		{	/* scalaire -> passage par valeur */
			if( type == TBYTE || type == TCHAR )
			{	/* char promu en int */
				my_va_arg(&in, ap, sizeof(int)) ;
				by = in ;
				memcpy(	(void *) (p+s[i].offset),
					(void *) &by , sizeof(by)) ;
			}
			else if (type == TSHORT )
			{	/* short promu en int */
				my_va_arg(&in, ap, sizeof(int)) ;
				sh = in ;
				memcpy(	(void *) (p+s[i].offset),
					(void *) &sh , sizeof(sh)) ;
			}
			else if (type == TFLOAT )
			{	/* float promu en double */
				my_va_arg(&dou, ap, sizeof(double)) ;
				flo = dou ;
				memcpy(	(void *) (p+s[i].offset),
					(void *) &flo , sizeof(flo)) ;
			}
			else if (type == TDOUBLE || 
				type == TLONG || type == TBLOC )
			{	/* 
				** Donnees passe dans la pile SANS promotion:
				** une simple copie octet par octet suffit
				*/
				my_va_arg((p+s[i].offset), ap, s[i].size) ;
			}
		}
		else
		{	
			char *q ;
			/* ce sont des tableaux => passage par adresse */
			my_va_arg( &q , ap , sizeof(q)) ;
		/*	printf("\n On a recupere [%s]", q) ; */
			/* traitement du cas particulier du tableau de char 
			** on respecte la semantique du char * du langage C
			** ie: tableau bornes par un 0
			*/
			if( type == TCHAR )
			{
				char *qq ;
				int si ;
				/* Note: strncpy tronque la chaine source 
				** si elle est trop longue, et complete
				** par des 0 si elle est trop courte
				*/
				strncpy( qq = p+s[i].offset, q, si = s[i].size ) ;
				/* si trop longue, le 0 de queue peut avoir
				** ete supprime, le remettre (en ecrasant)
				** le dernier caractere 
				*/
				qq[si-1] = 0 ;
			}
			else
			{
				memcpy((void *) (p+s[i].offset) , 
					(void *)q , s[i].size ) ;
			}
		}
	}
	va_end(ap) ;
}  
/* --------------------------------------------------------------------
**	Affiche le contenu d'un selecteur (DEBUG only)
*/
void PrintSelector( SELECTOR *s)
{
	int nbsel , nbelem, type , i;
		
	nbsel = s[0].size ;
	printf("\n ------ Table de selecteurs : (%p)",s) ;
	printf("\n nb d'entrees : %ld, type du tuple : [%s], Magic : %X",
		nbsel, ((SELECTOR0 *)s)->type_id->name, ((SELECTOR0 *)s)->magic ) ; 
		 
	
	for( i = 1 ;  i <= nbsel ; ++i )
	{
		/* calculer le nb d'elements, et le type */
		nbelem = s[i].nbelem & 0x1FFFFFFFL ; 
			/* masquer 3 bits de poids fort */
		type = s[i].nbelem >> 29 ;	
			/* unsigned -> pas de pb de bit de signe */
		printf("\n size = %ld, offset = %ld, arg_offset = %ld, nbelem = %d, type = %d",
			s[i].size, s[i].offset , s[i].arg_offset, nbelem, type) ;
					
	}		
}
#endif

#ifndef LNDB
/* ------------------------------------------------------------------
** Fonctions de service
**
*/
void *FillBuff(void *p, void *q, size_t s)
{
	memcpy(p,q,s) ;
	return (void *)(((char *)p)+s) ;
}
void *EmptyBuff(void *p, void *q, size_t s)
{
	memcpy(q,p,s) ;
	return (void *)(((char *)p)+s) ;
}
/* ------------------------------------------------------------------
**	Fonction Eval()
**	Retour: >  0	0K
**		== 0	Probleme d'allocation lors de l'eval
**		<  0	Erreur Linda (voir erreurs de lput() )
*/	

/* *****************************************************************
** Format du message d'evaluation:
** Il est compose d'un ensemble de TYPES (au sens de T-Linda)
** et d'un bloc de donnees, semblable a celui qui est transmis par out()
** Les types sont connus par leur Global_id, (adr sur serveur gerant 
** ce type) et leur description (ce qui permet au process EVALue de 
** recreer ce type en local, comme si il avait ete declare par 
** DefineTupleType() )
** Voici la structure du bloc memoire passe au serveur de processus:
** 
** 				Taille pour cette implementation (octets)
**----------------------------------------------------------------------
** Nombre de types		sizeof(int)
** Global_Id du type 1 		sizeof(TUPLE_TYPE *)
** Nom du Type			MAX_TPISIZE
** Descriptif du type		<variable, termine par '\0'>
** Filler, pour alignement	0 a (sizeof(TUPLE_TYPE *)-1) octets
**
** Global_Id du type 2 		sizeof(TUPLE_TYPE *)
** Nom du Type			MAX_TPISIZE
** Descriptif du type		<variable, termine par '\0'>
** Filler, pour alignement	0 a (sizeof(TUPLE_TYPE *)-1) octets
** 
** Etc...(autres types)
**
** Bloc de donnees		datasiz
*************************************************************************
*/
int Eval(char *procname, TUPLE_TYPE **tt, int nt, void *data, long datasiz)
{
	int i,k ;
	char name[MAX_TPISIZE], *p, *buff ;
	long totalsiz ;
	
	/* fabriquer le nom du tuple niveau 1 a generer (_XXXX) */
	name[0] = '_' ;
	strncpy(name+1, procname, MAX_TPISIZE-1) ;
	name[MAX_TPISIZE-1] = 0 ;
	
	/* calcul de la taille du buffer necessaire */
	totalsiz = sizeof(int) ;
	for(i = 0 ; i < nt ; ++i)
	{
		int fmtsiz ;
		
		totalsiz += (sizeof(TUPLE_TYPE *) +	/* global_id */
				MAX_TPISIZE ) ;		/* nom type */
		fmtsiz = strlen(tt[i]->fmt) + 1 ;	/* format + '\0' */
		fmtsiz = 
			(long) IncTo((unsigned long)fmtsiz , 
			       (unsigned long) sizeof(TUPLE_TYPE *) ) ; /* filler */
		totalsiz += fmtsiz ;
	}
	totalsiz += datasiz  ; 				/* donnees */
	
	p = buff = malloc( totalsiz ) ;
	if( p == NULL ) return 0 ;
	
	/* remplissage */
	p = FillBuff(p, &nt, sizeof(int) ) ;	/* nb de types */
	for( i = 0 ; i < nt ; ++i)
	{
		p = FillBuff(p, &(tt[i]->glob_handle), sizeof(TUPLE_TYPE *)) ;
		p = FillBuff(p, (tt[i]->name), (size_t)MAX_TPISIZE) ;
		k = strlen(tt[i]->fmt) + 1 ;
		FillBuff(p, (tt[i]->fmt), (size_t)k) ;
		p += IncTo(k , sizeof(TUPLE_TYPE *)) ; /* alignement */
	}
	if( data != NULL && datasiz != 0)
		p = FillBuff(p, data, (size_t) datasiz ) ;
	
	/* lancer l'eval */

	k = lput(name, buff, totalsiz ) ;
	free(buff) ;
	return k ;
}
#endif
	
		


#ifndef LNDB
/* -----------------------------------------------------------------------
** Les fonctions qui suivent permettent d'acrire des serveurs de 
** processus, qui implementent la fonction Linda eval()
** 
** Le principe de ces serveurs est le suivant:
** Chaque serveur de processus est un programme Linda Niveau 1
** (dans le sens ou il UTILISE les primitives Linda Niv. 1)
** qui attend des ordres d'activation de processus. 
** Un processus est activable si il est declare comme tel au serveur 
** (fonction RegisterProcess()).
** Ces processus sont simplement des fonctions C, "Forkee" par le
** serveur qui leur passe les parametres d'EVALuation.
** (ATTENTION: Le Fork d'Helios n'a rien a voir avec le fork Unix) 
** Ces parametres d'evaluation sont
**   -Un ensemble de types + le nb de types
**   -Les parametres de la fonction eval passes par le processus appelant +
**    la taille de ce bloc de donnees.
** Le prototype d'une fonction evaluee est donc:
**
** void evalfunction( TupleType *t, int tsize, void *data, long datasize ) ;
** (tsize est un nombre de types, datasize est un nombre d'octets)
*/

/* -----------------------------------------------------------------
** Ajoute un process a la liste des process Evaluables
** Retour: 1 si OK,   0 sinon.
*/
int RegisterProcess(	long stack_size ,
			char *name ,
			VoidFnPtr pf ,
			int flag )
{
	EVAL_PROCESS *p = malloc(sizeof(EVAL_PROCESS) ) ;
	
	if( p == NULL) return 0 ;
	
	p->stack_size = stack_size ;
	/* le nom d'enregistrement est de la forme "_nom" */
	p->name[0] = '_' ;
	strncpy(p->name+1, name, MAX_TPISIZE-1) ;
	p->name[MAX_TPISIZE-1] = 0 ; /* gestion troncature */
	p->pf = pf ;
	p->flag = flag ;
	
	/* maintenant , on lance un processus gardien pour ce process */
	if( Fork((WORD)1024 , &eval_guard, sizeof(p), p) == 0) return 0 ;
	return 1 ; 
}

/* -----------------------------------------------------------------
** Processus Gardien pour activer les processus evaluables
*/
void eval_guard( EVAL_PROCESS *pf )
{
	int i,nt ;
	char *buffer, *p  ;
	long datasize ;
	long n ;
	TUPLE_TYPE **tt, **ttsav,  *glob_handle , *loc ;
	char name[MAX_TPISIZE] ;
	int *pi ;
		
	for( ;; )
	{
		/* NOTE: lgetu() fait un realloc() 
		** => buffer doit etre NULL , si on veux eviter de recuperer
		** le meme bloc memoire pour chaque processus cree
		*/
		buffer = NULL ;
		n = lgetu(pf->name, (void **)&buffer) ;
	
		/* Un ordre Eval() a ete recu 
		** 1) recuperer les types globaux et leur description 
		** 2) creer les types locaux correspondants 
		** 3) fabriquer une table de types locaux 
		** 4) forker, ou appeler le processus, en lui passant cette table 
		** 
		** La methode retenue est de constituer la table
		** des types DANS le bloc memoire 'buffer' alloue
		** par lgetu(). Cette table est necessairement + courte
		** que les descriptifs de type recus, ce qui garantit le
		** non-debordement. Cela permet en outre de garder toutes
		** les info necessaires au processus forke
		** (nb de type, tab des types, datas) dans le MEME bloc
		** memoire, ce qui facilite la de-allocation. Cela
		** evite en outre des allocations supplementaires, qui
		** fragmentent la memoire.
		** Le bloc passe au processus a la structure suivante:
		** 1-Nb de types (nt, qui est un int)
		** 2-Magic number MAGIC_TYPE_TABLE pour authentifier le tout
		** 3-Local ID 1
		** 4-Local ID 2
		** ...
		** 5 Local ID nt
		** 6 (de la memoire inutilisee)
		** 7 Les donnees.
		*/ 
		
		p = buffer ;
		/* nb de types */
		tt = (TUPLE_TYPE **)(p = EmptyBuff( p , &nt, sizeof(nt))) ;
		
		/* recuperer nt fois les sequences
		** global_handler	
		** nom du type
		** description du type + filler
		*/	
		for( i = 0 ; i < nt ; ++i)
		{
			p = EmptyBuff(p, &glob_handle, sizeof( glob_handle) ) ;
			p = EmptyBuff(p, name, MAX_TPISIZE ) ;
			if( i == 0)
			{
				*tt++ = (TUPLE_TYPE *)MAGIC_TYPE_TABLE ;
				ttsav = tt ;
			}
			/* fabriquer le type 'herite' */

			loc = NewTupleType() ;
			if (loc == NULL)
				break;
			/* construction de l'arbre lie a ce type */
			if (!analyse_fmt(p, &(loc->decl)))
			{
				free(loc);
				loc = NULL;
				break;
			}
			/* stockage du nom */
			memcpy(loc->name, name, MAX_TPISIZE);
			/* du format */
			loc->fmt = Strdup(p);
			if (loc->fmt == NULL)
			{
				free(loc);
				loc = NULL;
				break;
			}

			loc->glob_handle = glob_handle ; /* initialiser le handle global */
			loc->numdef = 0; /* 0 indique un type herite */
			loc->next = NULL;
			/* chainer le type en fin de table */
			Wait(&_lndtypelock) ;
			*FindLocalType( NULL , &type_table ) = loc ;
			Signal(&_lndtypelock) ;

			/* stocker le type dans la table */
			*tt++ = loc ;
			
			/* sauter le descriptif de type et le '\0'*/
			while( *p++ ) ;
			
			/* sauter le filler */
			p = (char *)IncTo((unsigned long) p , 
			    (unsigned long)sizeof( TUPLE_TYPE *)) ;
		}

		/* ici, p pointe sur le debut du bloc de donnees:
		** on peut donc calculer la longueur de ce bloc 
		*/
		datasize = n - (p-buffer) ;
		/* activation du processus */
		if( EVL_MULTI & pf->flag)
		{
			pi = (int *) p ;
			if( 
			Fork(pf->stack_size, pf->pf, 	sizeof(ttsav) + 
							sizeof(nt) +
							sizeof(p) +
							sizeof(datasize) ,
					ttsav, nt,  p, datasize ) == 0 )
			
			{
				printf("\nevalguard: Unable to fork a new process\n") ;
			}
		}
		else
		{
			(*(pf->pf))(ttsav, nt, p, datasize) ;
		}

	} 	
					
}

/* -------------------------------------------------------------------
** 
** Cette fonction DOIT etre appelee a la fin de chaque processus EVALue
** Elle permet de 
** 1 - Detruire les types herites du processus appelant
** 2 - Desallouer le bloc de parametres passe au processus
** Retour: 1 si OK, 0 si parametre incorrect.
** Attention: On doit passer a cette fonction la table des types 
** telle qu'elle a ete passee au processus EVALue 
** (et surtout pas une copie de celle-ci, car la memoire contigue
** a cette table contient des informations vitales)
*/
int EndEval( TUPLE_TYPE **t)
{
	int *nbt = ((int *)(t-1)) - 1 ;
	int i ;
	
	if( t[-1] != (TUPLE_TYPE *)MAGIC_TYPE_TABLE ) 
	{
		return 0 ;
	}

	t[-1] = NULL ; /* invalidation au cas ou ce bloc serait reutilise */	
	Wait(&_lndtypelock) ;
	
	for( i = 0 ; i < *nbt ; ++i)
	{
		if( DeleteLocalType(t[i], &type_table) == NULL)
		{
			printf("\nEndEval: Corrupted Type Table\n") ;
			break ;
		} 
	}
	
	Signal(&_lndtypelock) ;
	
	free(nbt ) ;
	return 1 ;
}	
#endif	
#ifndef LNDB
/* ---------------------------------------------------------------------
**	Imprime la table des types (debug)
*/
void      print_types(TUPLE_TYPE * t)
{
	int       k;

	for (; t != NULL ; t = t->next )
	{
		k = strlen(t->fmt);
		if (k < 50)
		{
			printf("\nType <%s> - Decl : %ld - Descriptor : \"%s\"", 
				t->name , t->numdef, t->fmt);
			printf("\nGlob Id=%p Loc Id=%p", t->glob_handle, t) ;
		}
		else
		{
			int       c;

			printf("\nType <%s> - Decl : %ld - Descriptor : %d+1 bytes", 
				t->name, t->numdef , k);
			c = t->fmt[24];
			t->fmt[24] = 0;
			printf("\n\"%s...", t->fmt);
			t->fmt[24] = c;
			printf("%s\"", t->fmt + k - 25);
			printf("\nGlob Id=%p Loc Id=%p", t->glob_handle, t) ;
		}
		fflush(stdout);
	}
}

/* --------------------------------------------------------------------
	Renvoie la valeur x , incrementee pour etre un 
	multiple de k
*/

unsigned long  IncTo( unsigned long x, unsigned long k )
{
	unsigned long q = x % k;
	if( q != 0 )
		x += (k-q) ;
	return x ;
} 

#endif

