#include <stdio.h>
#include "linda3.h"

typedef struct barriere
{
	char nom[40] ;
	long num ;
} BARRIERE ;

/* Prototype des "fonctions-processus" */
void bars( TupleType *t, int tsize, int *data, long datasize) ;
void cnk(TupleType *tt, int ntt, long *nk, long nksz) ;
void scycle(TupleType *tt, int ntt, int *param, long paramsiz) ;


/* Exemple 1 --------------------------------------------------------
** Synchronisation par barriere:
** voici le processus a synchroniser
*/
void bars( TupleType *t, int tsize, int *data, long datasize)
{

	BARRIERE b,i_go ;
	TupleType barr 		= t[0] ;
	TupleSelector sel 	= CreateTupleSelector( barr , "{.?}" ) ;
	TupleSelector sel1 	= CreateTupleSelector( barr , "{..}" ) ;
	int k ;	

	datasize = tsize = *data ;	/* pour le compilo */		
	MakeByteArray(b.nom,"BARRIER") ;
	MakeByteArray(i_go.nom, "I-GO" ) ;	

	in(sel, &b) ;
	b.num-- ;
	k = out(barr, &b ) ;
	b.num = 0 ;
	rd(sel1,&b) ;
	k = out(barr, &i_go ) ;
	
	
	/* nettoyage */
	free(sel ) ; free(sel1 ) ;
	EndEval(t) ;
}
	
/* Exemple 2 -----------------------------------------------------------
** Processus de calcul de CNK
** cnk(n,k) = n! / ( k! * (n-k)! )
*/
void cnk(TupleType *tt, int ntt, long *nk, long nksz)
{
    struct CNK { long n,k,cnk ; } tuple ;
    long i, imin ;
    double  num , den ;

    ntt = nksz  ;		/* pour le compilo */
    
    tuple.n = nk[0], tuple.k = nk[1];
    for(num = 1, i=tuple.n, imin=i-tuple.k ;                    
                    i > imin ; --i)
         num *= i ;
    for(den=1, i=tuple.k ; i > 1 ; --i)
         den *= i ;
    tuple.cnk = (long) (num/den) ;

    /* depot du tuple-resultat */
    out(tt[0], &tuple) ;


    /* terminaison du processus */
    EndEval(tt) ;
}
/*  Exemple 3 ********************************************************
**  Simulation de cycles de calcul de radiosite
**  ATTENTION T-Linda V3.5 ou posterieure 
**  (utilise les declarations a champs variables)
*/	

/********** definition des nouveaux types */

typedef struct				/* paquet de donnees */
{
  char d[8] ;				/* donnees de calcul */
  short c[1] ;				/* compteurs de passage */
} PAQUET ;

/********** fonctions annexes */
PAQUET *palloc(int nbprocessus) ;

/* Allocation d'un paquet dont la taille 
** est fonction du nombre de processus
*/
PAQUET *palloc(int nbprocessus)
{
  PAQUET *paquet;

  paquet = (PAQUET *)malloc( sizeof(PAQUET) + (nbprocessus-1)*sizeof(short) );
  if (paquet == NULL)
    {
SCREEN ;
      fprintf(stderr, "pas assez de memoire pour creer un paquet\n") ;
NOSCREEN ;
    }
  return paquet;
}

/********** Fonction-processus */

void scycle( TupleType *tt, int ntt, int *par, long parsize) 
{
  int np = par[0] ;			/* nombre de paquets de donnees */
  int ne = par[1] ;			/* nombre de processus de travail */
  int nc = par[2] ;			/* nombre de cycles de calcul */
  int nn = par[3] ;			/* numero d'ordre du processus */
  PAQUET *p;				/* un paquet de donnees */
  TupleSelector ts;			/* selecteur pour les typles "paquet" */
  int i, j, k;				/* quelque variables de travail */

  ntt = parsize ;			/* pour le compilo (WAR var non util.) */
  
  /**** initialisations ****/

  p = palloc(ne);			/* alloue la memoire pour un paquet */

  ts = CreateTupleSelector(tt[0], 
  	"{? %d? . %d?}", nn, ne-nn-1);	/* cree un selecteur {? ???.????} */

  /**** boucle de calcul ****/

  for (i = 0; i < nc; ++i)		/* nombre de cycles */
    {
      for (j = 0; j < np; ++j)		/* nombre de paquets */
	{
	  p->c[nn] = i;			/* lire un paquet du cycle courant */
	  k = in(ts, p);		/* retrait d'un paquet du T.S. */
SCREEN ;
	  printf("(%d) lecture paquet %s : %d ", nn, p->d, k);

	  printf("compteurs :");
	  for (k=0; k < ne; ++k)
	    printf(" %d", p->c[k]);
	  printf("\n");
NOSCREEN ;
	  p->c[nn] = i+1;		/* incrementer le compteur de passage */
	  k = out(tt[0], p);		/* depose le paquet dans le T.S. */
SCREEN ;
	  printf("(%d) ecriture du paquet %s : %d\n", nn, p->d, k);
NOSCREEN ;
	}
    }

  /**** terminaisons ****/

  free(p);		/* destruction du paquet */
  free(ts) ;		/* destruction du selecteur */
  EndEval(tt) ; 	/* terminaison et nettoyage environnement */
  
}


/* -------------------------------------------------------------
** Serveur proprement dit
*/

int main()
{
	void *dummy ;
	
	if( RegisterProcess(1024L, "bars", &bars, EVL_MULTI ) )
	{
		SCREEN ;
		printf("\n Process <bars> enregistre") ;
		NOSCREEN ;
	}
	else
	{
		SCREEN ;
		printf("\n Echec de l'enregistrement du process <bars>") ;
		NOSCREEN ;
	}
	if( RegisterProcess(1024L, "CNK", &cnk, EVL_MULTI ) )
	{
		SCREEN ;
		printf("\n Process <cnk> enregistre") ;
		NOSCREEN ;
	}
	else
	{
		SCREEN ;
		printf("\n Echec de l'enregistrement du process <cnk>") ;
		NOSCREEN ;
	}
	if( RegisterProcess(1024L, "scycle", &scycle, EVL_MULTI ) )
	{
		SCREEN ;
		printf("\n Process <scycle> enregistre") ;
		NOSCREEN ;
	}
	else
	{
		SCREEN ;
		printf("\n Echec de l'enregistrement du process <scycle>") ;
		NOSCREEN ;
	}

	SCREEN ;
	printf("\n This process server can be stopped"
		"\n by a level 1 STOPPROCESS tuple") ;
	NOSCREEN ;
	
	/* attente de l'arret */
	lgetu("STOPPROCESS", &dummy) ;
	
	SCREEN ;
	printf("\n Process server killed by STOPPROCESS request.\n") ;
	NOSCREEN ;
	return 0 ;
	
}

